#!/usr/bin/perl -w
######################################################################
#
# $Id: WebJob-KvpRoutines.t,v 1.8 2008/01/12 17:37:53 mavrik Exp $
#
######################################################################
#
# Copyright 2007-2008 The WebJob Project, All Rights Reserved.
#
######################################################################
#
# Purpose: Tests for WebJob::KvpRoutines.
#
######################################################################

use 5.008;
use strict;
use Test;

BEGIN
{
  my %hPlan =
  (
    'tests' => 6,
  );
  plan(%hPlan);
};

######################################################################
#
# Tests
#
######################################################################

  ####################################################################
  #
  # Test: The target module must load without error.
  #
  ####################################################################

  use WebJob::KvpRoutines;
  ok(1);

  ####################################################################
  #
  # Test: Parse a simple config file.
  #
  ####################################################################

  my $sFile = "cfg";
  my %hProperties = ();
  if (open(FH, "> $sFile"))
  {
    print FH <<EOF;

# Comment starting at column zero.
 # Comment starting at column one.
# Alpha
a_a=a
a_b =b
a_c = c# Space before the value.
a_d = d # Space before and after the value.
 a_e=e
 a_f =f# Space before the value.
 a_g = g
 a_h = h # Space before and after the value.
# Digit
d_a=1
d_b =2
d_c = 3# Space before the value.
d_d = 4 # Space before and after the value.
 d_e=5
 d_f =6# Space before the value.
 d_g = 7
 d_h = 8 # Space before and after the value.
# Empty
e_a=
e_b =
e_c = # Space before the value.
e_d =  # Space and comment after the value.
 e_e=
 e_f =
 e_g = # Space before the value.
 e_h =  # Space and comment after the value.
# Other
o_a=foo
o_b=foo123
o_c=123
O_A=bar    # Case should be ignored for keys, and repeats are allowed.
O_B=bar456 # Case should be ignored for keys, and repeats are allowed.
O_C=456    # Case should be ignored for keys, and repeats are allowed.
# Boolean
b_a=Y
b_b=y
b_c=N
b_d=n
b_e=T
b_f=t
b_g=F
b_h=f
b_i=1
b_j=0
b_k=YES
b_l=Yes
b_m=yes
b_n=NO
b_o=No
b_p=no
b_q=TRUE
b_r=True
b_s=true
b_t=FALSE
b_u=False
b_v=false
# Comment (Note: The comments below are spaced so they line up in the config file.)
c_a=1#        # This value is supposed to be '1'.
c_b=1#2#      # This value is supposed to be '1'.
c_c=1#2#3#    # This value is supposed to be '1'.
c_d=1\\#       # This value is supposed to be '1#'.
c_e=1\\#2\\#    # This value is supposed to be '1#2#'.
c_f=1\\#2\\#3\\# # This value is supposed to be '1#2#3#'.
c_g=1\\\\#      # This value is supposed to be '1\\#'.
EOF
    close(FH);
  }
  my %hLArgs =
  (
    'File'           => $sFile,
    'Properties'     => \%hProperties,
    'RequireAllKeys' => 1,
    'Template'       =>
    {
      # Alpha
      'a_a'          => "[A-Za-z]",
      'a_b'          => "[A-Za-z]",
      'a_c'          => "[A-Za-z]",
      'a_d'          => "[A-Za-z]",
      'a_e'          => "[A-Za-z]",
      'a_f'          => "[A-Za-z]",
      'a_g'          => "[A-Za-z]",
      'a_h'          => "[A-Za-z]",
      # Boolean
      'b_a'          => "[10YyNnTtFf]",
      'b_b'          => "[10YyNnTtFf]",
      'b_c'          => "[10YyNnTtFf]",
      'b_d'          => "[10YyNnTtFf]",
      'b_e'          => "[10YyNnTtFf]",
      'b_f'          => "[10YyNnTtFf]",
      'b_g'          => "[10YyNnTtFf]",
      'b_h'          => "[10YyNnTtFf]",
      'b_i'          => "[10YyNnTtFf]",
      'b_j'          => "[10YyNnTtFf]",
      'b_k'          => "(?i)(?:false|no|true|yes|no)",
      'b_l'          => "(?i)(?:false|no|true|yes|no)",
      'b_m'          => "(?i)(?:false|no|true|yes|no)",
      'b_n'          => "(?i)(?:false|no|true|yes|no)",
      'b_o'          => "(?i)(?:false|no|true|yes|no)",
      'b_p'          => "(?i)(?:false|no|true|yes|no)",
      'b_q'          => "(?i)(?:false|no|true|yes|no)",
      'b_r'          => "(?i)(?:false|no|true|yes|no)",
      'b_s'          => "(?i)(?:false|no|true|yes|no)",
      'b_t'          => "(?i)(?:false|no|true|yes|no)",
      'b_u'          => "(?i)(?:false|no|true|yes|no)",
      'b_v'          => "(?i)(?:false|no|true|yes|no)",
      # Comment
      'c_a'          => "1",
      'c_b'          => "1",
      'c_c'          => "1",
      'c_d'          => "1#",
      'c_e'          => "1#2#",
      'c_f'          => "1#2#3#",
      'c_g'          => "1\\\\#", # This is supposed to match the literal value '1\#'.
      # Digit
      'd_a'          => "\\d",
      'd_b'          => "\\d",
      'd_c'          => "\\d",
      'd_d'          => "\\d",
      'd_e'          => "\\d",
      'd_f'          => "\\d",
      'd_g'          => "\\d",
      'd_h'          => "\\d",
      # Empty
      'e_a'          => "",
      'e_b'          => "",
      'e_c'          => "",
      'e_d'          => "",
      'e_e'          => "",
      'e_f'          => "",
      'e_g'          => "",
      'e_h'          => "",
      # Other
      'o_a'          => "[\\w-]+",
      'o_b'          => "[\\w\\d-]+",
      'o_c'          => "\\d+",
    },
    'VerifyValues'   => 1,
  );
  ok(KvpGetKvps(\%hLArgs));
# print STDERR "\n";
# foreach my $sKey (sort(keys(%hProperties)))
# {
#   print STDERR "$sKey=[$hProperties{$sKey}]\n";
# }

  ####################################################################
  #
  # Test: Case sensitive keys.
  #
  ####################################################################

  %hProperties = ();
  %hLArgs =
  (
    'File'           => $sFile,
    'MatchKeyCase'   => 1,
    'Properties'     => \%hProperties,
    'RequireAllKeys' => 1,
    'Template'       =>
    {
      'o_a'          => "[\\w-]+",
      'o_b'          => "[\\w\\d-]+",
      'o_c'          => "\\d+",
      'O_A'          => "[\\w-]+",
      'O_B'          => "[\\w\\d-]+",
      'O_C'          => "\\d+",
    },
    'VerifyValues'   => 1,
  );
  ok(KvpGetKvps(\%hLArgs));
# print STDERR "\n";
# foreach my $sKey (sort(keys(%hProperties)))
# {
#   print STDERR "$sKey=[$hProperties{$sKey}]\n";
# }

  ####################################################################
  #
  # Test: Regular expression keys.
  #
  ####################################################################

  %hProperties = ();
  %hLArgs =
  (
    'File'           => $sFile,
    'MatchKeyCase'   => 1, # This option should set for regular expression keys, and the expression should deal with case.
    'Properties'     => \%hProperties,
    'RequireAllKeys' => 0, # This option should not be set for regular expression keys.
    'Template'       =>
    {
      'a_[a-h]'      => "[A-Za-z]",
      'b_[a-v]'      => "(?i)(?:false|no|true|yes|no|[10YNTF])",
      'c_[a-g]'      => "(?:1(?:#?(?:2#?(?:3#?)?)?)?|1\\\\#)",
      'd_[a-h]'      => "\\d",
      'e_[a-h]'      => "",
      'o_[a-c]'  => "[\\w\\d-]+", # This expression purposely excludes the 'O_[A-C]' keys.
    },
    'VerifyValues'   => 1,
  );
  ok(KvpGetKvps(\%hLArgs));
# print STDERR "\n";
# foreach my $sKey (sort(keys(%hProperties)))
# {
#   print STDERR "$sKey=[$hProperties{$sKey}]\n";
# }

  ####################################################################
  #
  # Test: Write a simple config file.
  #
  ####################################################################

  %hLArgs =
  (
    'Delimiter'      => "=",
    'File'           => $sFile,
    'Properties'     => \%hProperties,
    'Template'       =>
    {
      'a_[a-h]'      => "[A-Za-z]",
      'b_[a-v]'      => "(?i)(?:false|no|true|yes|no|[10YNTF])",
      'c_[a-g]'      => "(?:1(?:#?(?:2#?(?:3#?)?)?)?|1\\\\#)",
      'd_[a-h]'      => "\\d",
      'e_[a-h]'      => "",
      'o_[a-c]'  => "[\\w\\d-]+", # This expression purposely excludes the 'O_[A-C]' keys.
    },
    'VerifyValues'   => 1,
  );
  ok(KvpSetKvps(\%hLArgs));

  ####################################################################
  #
  # Test: Append to a config file.
  #
  ####################################################################

  %hLArgs =
  (
    'AppendToFile'   => 1,
    'Delimiter'      => "=",
    'File'           => $sFile,
    'Properties'     =>
    {
      'O_A'          => "bar",
      'O_B'          => "bar456",
      'O_C'          => "456",
    },
    'Template'       =>
    {
      'O_[A-C]'      => "[\\w\\d-]+",
    },
    'VerifyValues'   => 1,
  );
  ok(KvpSetKvps(\%hLArgs));

  unlink($sFile);

