#!/usr/bin/perl -w
######################################################################
#
# $Id: WebJob-LogRoutines.t,v 1.2 2008/08/08 19:05:55 mavrik Exp $
#
######################################################################
#
# Copyright 2008-2008 The WebJob Project, All Rights Reserved.
#
######################################################################
#
# Purpose: Tests for WebJob::LogRoutines.
#
######################################################################

use 5.008;
use strict;
use Test;
use Digest::MD5;

BEGIN
{
  my %hPlan =
  (
    'tests' => 258,
  );
  plan(%hPlan);
};

######################################################################
#
# Tests
#
######################################################################

  ####################################################################
  #
  # Test: The target module must load without error.
  #
  ####################################################################

  use WebJob::LogRoutines;
  ok(1);

  ####################################################################
  #
  # Test: Generate a simple log message.
  #
  ####################################################################

  my $sLogFile = "log";

  my %hOptions =
  (
    'LogEpoch'       => 0,
    'LogFields'      => ["Field1", "Field2", "Message"],
    'LogFile'        => $sLogFile,
    'LogValues'      => {},
    'RevertToStderr' => 1,
    'UseGmt'         => 1,
  );

  foreach my $sField1 ("field_1", "field 1", "field\t1", "field\n1", "field\r1", "field+1", "field%1", undef)
  {
    $hOptions{'LogValues'}{'Field1'} = $sField1;
    foreach my $sField2 ("field_2", "field 2", "field\t2", "field\n2", "field\r2", "field+2", "field%2", undef)
    {
      $hOptions{'LogValues'}{'Field2'} = $sField2;
      foreach my $sMessage ("pass", "fail", "pass and/or fail", undef)
      {
        $hOptions{'LogValues'}{'Message'} = $sMessage;
        ok(LogNf1vMessage(\%hOptions));
        $hOptions{'LogEpoch'}++;
      }
    }
  }

  ####################################################################
  #
  # Test: Check the hash of the log file.
  #
  ####################################################################

  my $oMd5 = Digest::MD5->new;
  open(FH, "< $sLogFile");
  $oMd5->addfile(*FH);
  my $sMd5 = $oMd5->hexdigest;
  close(FH);
  ok($sMd5 eq "80530eeaaa145a13fff8cbb5a53f57c7");

  ####################################################################
  #
  # Cleanup and go home.
  #
  ####################################################################

  unlink($sLogFile);

