#!/usr/bin/perl -w
######################################################################
#
# $Id: WebJob-TimeRoutines.t,v 1.4 2008/08/06 01:08:33 mavrik Exp $
#
######################################################################
#
# Copyright 2007-2008 The WebJob Project, All Rights Reserved.
#
######################################################################
#
# Purpose: Tests for WebJob::TimeRoutines.
#
######################################################################

use 5.008;
use strict;
use Test;

BEGIN
{
  my %hPlan =
  (
    'tests' => 18,
  );
  plan(%hPlan);
};

######################################################################
#
# Tests
#
######################################################################

  ####################################################################
  #
  # Test: The target module must load without error.
  #
  ####################################################################

  my ($sDateTime, $sDhms, $sSeconds);

  use WebJob::TimeRoutines;
  ok(1);

  ####################################################################
  #
  # Test: The UNIX epoch yields 0.
  #
  ####################################################################

  $sSeconds = DateTimeToSeconds("1970-01-01 00:00:00", 1);
  ok((defined($sSeconds) && $sSeconds == 0));

  ####################################################################
  #
  # Test: 0 yields the UNIX epoch.
  #
  ####################################################################

  $sDateTime = SecondsToDateTime(0, 1);
  ok((defined($sDateTime) && $sDateTime =~ /^1970-01-01 00:00:00$/));

  ####################################################################
  #
  # Test: undef yields "?".
  #
  ####################################################################

  $sDhms = SecondsToDhms(undef);
  ok((defined($sDhms) && $sDhms =~ /^\?$/));

  ####################################################################
  #
  # Test: Various integer inputs yield expected results.
  #
  ####################################################################

  my @aPairs =
  (
    "-2:?",
    "-1:?",
    "0:0s",
    "1:1s",
    "2:2s",
    "60:1m0s",
    "61:1m1s",
    "3600:1h0m0s",
    "3660:1h1m0s",
    "3661:1h1m1s",
    "86400:1d0h0m0s",
    "90000:1d1h0m0s",
    "90060:1d1h1m0s",
    "90061:1d1h1m1s",
  );

  foreach my $sPair (@aPairs)
  {
    my ($sSeconds, $sExpectedResult) = split(/:/, $sPair);
    $sDhms = SecondsToDhms($sSeconds);
    ok((defined($sDhms) && $sDhms =~ /^$sExpectedResult$/));
  }

