#!/usr/bin/perl -w
######################################################################
#
# $Id: nph-webjob.cgi,v 1.72 2008/12/17 03:54:49 mavrik Exp $
#
######################################################################
#
# Copyright 2001-2008 Klayton Monroe, All Rights Reserved.
#
######################################################################
#
# Purpose: Process webjob GET/JOB/PUT requests.
#
######################################################################

use strict;
use Fcntl qw(:flock);
use FindBin qw($Bin $RealBin); use lib ("$Bin/../../lib/perl5/site_perl", "$RealBin/../../lib/perl5/site_perl", "/usr/local/webjob/lib/perl5/site_perl");
use WebJob::JqdRoutines;
use WebJob::KvpRoutines;
use WebJob::LogRoutines;
use WebJob::ValidationRoutines;

######################################################################
#
# Main Routine
#
######################################################################

  my (%hProperties, %hReturnCodes, $sLocalError);

  %hReturnCodes =
  (
    '200' => "OK",
    '204' => "No Content",
    '206' => "Partial Content",
    '251' => "Link Test OK",
    '403' => "Forbidden",
    '404' => "Not Found",
    '405' => "Method Not Allowed",
    '450' => "Invalid Query",
    '451' => "File Already Exists",
    '452' => "Username Undefined",
    '453' => "Username-ClientId Mismatch",
    '454' => "Content-Length Undefined",
    '455' => "Content-Length Exceeds Limit",
    '456' => "Content-Length Mismatch",
    '457' => "File Not Available",
    '458' => "Invalid Protocol",
    '459' => "Payload Signature Not Available",
    '460' => "Queue Not Active",
    '461' => "Queue Not Available",
    '462' => "Queue Not Mapped",
    '470' => "CommonName Undefined",
    '471' => "CommonName-ClientId Mismatch",
    '500' => "Internal Server Error",
    '503' => "Service Unavailable",
    '550' => "Internal Server Initialization Error",
    '553' => "Service Disabled",
  );

  ####################################################################
  #
  # Punch in and go to work.
  #
  ####################################################################

  $hProperties{'StartTime'} = time();
  $hProperties{'Version'} = sprintf("%s %s", __FILE__, ('$Revision: 1.72 $' =~ /^.Revision: ([\d.]+)/));

  ####################################################################
  #
  # Create/Verify run time environment, and process GET/PUT requests.
  #
  ####################################################################

  if (!defined(CreateRunTimeEnvironment(\%hProperties, \$sLocalError)))
  {
    $hProperties{'ReturnStatus'} = 550;
    $hProperties{'ReturnReason'} = $hReturnCodes{$hProperties{'ReturnStatus'}};
    $hProperties{'ErrorMessage'} = $sLocalError;
  }
  else
  {
    if (Yes($hProperties{'SslRequireSsl'}) && (!defined($hProperties{'Https'}) || $hProperties{'Https'} !~ /^[Oo][Nn]$/))
    {
      $hProperties{'ReturnStatus'} = 458;
      $hProperties{'ReturnReason'} = $hReturnCodes{$hProperties{'ReturnStatus'}};
      $hProperties{'ErrorMessage'} = "HTTPS required, but client is speaking HTTP";
    }
    else
    {
      if ($hProperties{'RequestMethod'} eq "GET")
      {
        $hProperties{'ReturnStatus'} = ProcessGetRequest(\%hProperties, \$sLocalError);
        $hProperties{'ReturnReason'} = $hReturnCodes{$hProperties{'ReturnStatus'}};
        $hProperties{'ErrorMessage'} = $sLocalError;
        $hProperties{'ExpandTriggerCommandLineRoutine'} = \&ExpandGetTriggerCommandLine;
      }
      elsif ($hProperties{'RequestMethod'} eq "PUT")
      {
        $hProperties{'ReturnStatus'} = ProcessPutRequest(\%hProperties, \$sLocalError);
        $hProperties{'ReturnReason'} = $hReturnCodes{$hProperties{'ReturnStatus'}};
        $hProperties{'ErrorMessage'} = $sLocalError;
        $hProperties{'ExpandTriggerCommandLineRoutine'} = \&ExpandPutTriggerCommandLine;
      }
      else
      {
        $hProperties{'ReturnStatus'} = 405;
        $hProperties{'ReturnReason'} = $hReturnCodes{$hProperties{'ReturnStatus'}};
        $hProperties{'ErrorMessage'} = "Method ($hProperties{'RequestMethod'}) not allowed";
      }
    }
  }
  $hProperties{'ServerContentLength'} = SendResponse(\%hProperties);

  ####################################################################
  #
  # Conditionally log the GET/PUT transaction.
  #
  ####################################################################

  if (Yes($hProperties{'EnableLogging'}))
  {
    NphLogMessage(\%hProperties);
  }

  ####################################################################
  #
  # Conditionally pull the GET/PUT trigger and log the result.
  #
  ####################################################################

  if
  (
    $hProperties{'ReturnStatus'} == 200 &&
    (
      ($hProperties{'RequestMethod'} eq "GET" && Yes($hProperties{'GetTriggerEnable'})) ||
      ($hProperties{'RequestMethod'} eq "PUT" && Yes($hProperties{'PutTriggerEnable'}))
    )
  )
  {
    $hProperties{'TriggerEpoch'} = time();
    $hProperties{'TriggerPidLabel'} = "parent";
    $hProperties{'TriggerPid'} = $$;
    if (!defined(TriggerExecuteCommandLine(\%hProperties, \$sLocalError)))
    {
      $hProperties{'TriggerState'} = "failed";
      $hProperties{'TriggerMessage'} = $sLocalError;
      if (Yes($hProperties{'EnableLogging'}))
      {
        TriggerLogMessage(\%hProperties);
      }
    }
  }

  ####################################################################
  #
  # Clean up and go home.
  #
  ####################################################################

  1;


######################################################################
#
# CheckHostAccessList
#
######################################################################

sub CheckHostAccessList
{
  my ($sCidrList, $sIp) = @_;

  my $sCidrRegex = qq(((?:\\d{1,3}[.]){3}(?:\\d{1,3}))\/(\\d{1,2}));

  my $sIpRegex   = qq((?:\\d{1,3}[.]){3}(?:\\d{1,3}));

  if (!defined($sCidrList) || !defined($sIp))
  {
    return 0;
  }

  if ($sIp !~ /^$sIpRegex$/)
  {
    return 0;
  }

  $sCidrList =~ s/\s+//g; # Remove whitespace.

  foreach my $sCidr (split(/,/, $sCidrList))
  {
    next if ($sCidr !~ /^$sCidrRegex$/);
    my ($sNetwork, $sMaskBits) = ($1, $2);
    my $sBits2Clear = 32 - $sMaskBits - 1;
    my $sNetMask = 0xffffffff;
    foreach my $sBit (0..$sBits2Clear)
    {
      $sNetMask ^= 1 << $sBit;
    }
    my $sBinNetwork = hex(sprintf("%02x%02x%02x%02x", split(/\./, $sNetwork)));
    my $sBinNetmask = hex(sprintf("%08x", $sNetMask));
    $sBinNetwork &= $sBinNetmask; # Cleanup the network address -- in case the user did not.
    my $sBinIp = hex(sprintf("%02x%02x%02x%02x", split(/\./, $sIp)));
    if (($sBinIp & $sBinNetmask) == $sBinNetwork)
    {
      return 1; # We have a winner!
    }
  }

  return 0;
}


######################################################################
#
# CreateJobList
#
######################################################################

sub CreateJobList
{
  my ($phProperties, $psError) = @_;

  ####################################################################
  #
  # Make sure that required inputs are defined.
  #
  ####################################################################

  my %hLArgs =
  (
    'Hash' => $phProperties,
    'Keys' =>
    [
      'ClientId',
      'JobQueuePqActiveLimit',
      'JobQueuePqAnswerLimit',
      'JobQueueSqActiveLimit',
      'JobQueueSqAnswerLimit',
      'JqdJobCount',
      'JqdJobType',
      'JqdQueueDirectory',
      'JqdQueueName',
    ],
  );
  if (!defined(VerifyHashKeys(\%hLArgs)))
  {
    $$psError = $hLArgs{'Error'} if (defined($phProperties));
    return undef;
  }

  ####################################################################
  #
  # Initialize the job list.
  #
  ####################################################################

  $$phProperties{'JqdJobList'} = "";

  ####################################################################
  #
  # Lock the queue. This is an exclusive lock.
  #
  ####################################################################

  my (%hQueueLockArgs);

  %hQueueLockArgs =
  (
    'LockFile'  => $$phProperties{'JqdQueueDirectory'} . "/" . "change.lock",
  );

  if (!JqdLockFile(\%hQueueLockArgs))
  {
    $$psError = $hQueueLockArgs{'Error'};
    return undef;
  }

  ####################################################################
  #
  # Check to see if the queue has been frozen.
  #
  ####################################################################

  my ($sQueueIsFrozen);

  %hLArgs =
  (
    'Directory' => $$phProperties{'JqdQueueDirectory'},
  );

  $sQueueIsFrozen = JqdIsQueueFrozen(\%hLArgs);
  if (!defined($sQueueIsFrozen))
  {
    $$psError = $hLArgs{'Error'};
    JqdUnlockFile(\%hQueueLockArgs);
    return undef;
  }

  if ($sQueueIsFrozen)
  {
    $$psError = "Queue is frozen (check for freeze.lock)";
    JqdUnlockFile(\%hQueueLockArgs);
    return undef;
  }

#FIXME
  ####################################################################
  #
  # Determine the number of active (sent/open) jobs.
  #
  ####################################################################

  my ($sOpenJobCount, $sSentJobCount);

  %hLArgs =
  (
    'Directory'      => $$phProperties{'JqdQueueDirectory'},
    'JobCount'       => 0,
    'JobFiles'       => [],
    'JobState'       => "sent",
    'JobType'        => $$phProperties{'JqdJobType'},
    'ReturnJobFiles' => 0,
  );
  $sSentJobCount = JqdGetQueuedJobs(\%hLArgs);
  if (!defined($sSentJobCount))
  {
    $$psError = $hLArgs{'Error'};
    JqdUnlockFile(\%hQueueLockArgs);
    return undef;
  }

  $hLArgs{'JobState'} = "open";
  $sOpenJobCount = JqdGetQueuedJobs(\%hLArgs);
  if (!defined($sOpenJobCount))
  {
    $$psError = $hLArgs{'Error'};
    JqdUnlockFile(\%hQueueLockArgs);
    return undef;
  }

  ####################################################################
  #
  # Determine the number of jobs that we are willing to cut loose.
  #
  ####################################################################

  my ($sActiveCount, $sActiveLimit, $sAnswerLimit, $sCustomActiveLimit, $sCustomAnswerLimit);

  $sActiveCount = $sSentJobCount + $sOpenJobCount;

  if ($$phProperties{'JqdJobType'} eq "serial")
  {
    $sCustomActiveLimit = $$phProperties{'JobQueueSqActiveLimit'};
    $sCustomAnswerLimit = $$phProperties{'JobQueueSqAnswerLimit'};
  }
  else
  {
    $sCustomActiveLimit = $$phProperties{'JobQueuePqActiveLimit'};
    $sCustomAnswerLimit = $$phProperties{'JobQueuePqAnswerLimit'};
  }

  if ($sCustomActiveLimit == 0)
  {
    $sActiveLimit = 0; # Zero means there is no limit.
  }
  else
  {
    if ($sActiveCount >= $sCustomActiveLimit)
    {
      $$psError = substr($$phProperties{'JqdJobType'}, 0, 1) . "-queue is blocked with $sSentJobCount sent and $sOpenJobCount open jobs"; # This is put inside ()s by the caller.
      JqdUnlockFile(\%hQueueLockArgs);
      return "0/0"; # This case is not treated as an error.
    }
    $sActiveLimit = $sCustomActiveLimit - $sActiveCount;
  }

  if ($sCustomAnswerLimit == 0)
  {
    $sAnswerLimit = $sActiveLimit;
  }
  else
  {
    $sAnswerLimit = ($sActiveLimit < $sCustomAnswerLimit) ? $sActiveLimit : $sCustomAnswerLimit;
  }

  ####################################################################
  #
  # Get the current list of jobs, but don't pull more jobs than were
  # requested by the client or are allowed based on the limits
  # computed above.
  #
  ####################################################################

  my (@aTodoJobs, $sTodoJobCount, $sTodoJobLimit);

  if ($$phProperties{'JqdJobCount'} == 0)
  {
    if ($sAnswerLimit == 0)
    {
      $sTodoJobLimit = 0; # Zero means there is no limit.
    }
    else
    {
      $sTodoJobLimit = $sAnswerLimit;
    }
  }
  else
  {
    if ($sAnswerLimit == 0)
    {
      $sTodoJobLimit = $$phProperties{'JqdJobCount'};
    }
    elsif ($sAnswerLimit >= $$phProperties{'JqdJobCount'})
    {
      $sTodoJobLimit = $$phProperties{'JqdJobCount'};
    }
    else
    {
      $sTodoJobLimit = $sAnswerLimit;
    }
  }

  %hLArgs =
  (
    'Directory'      => $$phProperties{'JqdQueueDirectory'},
    'JobCount'       => $sTodoJobLimit,
    'JobFiles'       => \@aTodoJobs,
    'JobState'       => "todo",
    'JobType'        => $$phProperties{'JqdJobType'},
    'ReturnJobFiles' => 1,
  );
  $sTodoJobCount = JqdGetQueuedJobs(\%hLArgs);
  if (!defined($sTodoJobCount))
  {
    $$psError = $hLArgs{'Error'};
    JqdUnlockFile(\%hQueueLockArgs);
    return undef;
  }

  if ($sTodoJobCount == 0)
  {
    $$psError = substr($$phProperties{'JqdJobType'}, 0, 1) . "-queue is empty"; # This is put inside ()s by the caller.
    JqdUnlockFile(\%hQueueLockArgs);
    return "0/0"; # This case is not treated as an error.
  }

  ####################################################################
  #
  # Prepare the job list updating the queue state along the way. If
  # there's an error in the loop, stop processing and return any jobs
  # that successfully made it through the state transition. Check to
  # ensure that the job file is writable. This is done now because
  # state transitions later on need to insert additional content, so
  # it's better to fail here. This does two things: 1) it keeps the
  # job in the todo state and 2) it causes the queue to be frozen.
  #
  ####################################################################

  my (@aJobList, %hJobTemplate, $sSafeJobCount);

  %hJobTemplate =
  (
    'Command'      => $$phProperties{'CommonRegexes'}{'ClientSuppliedFilename'},
    'CommandAlias' => $$phProperties{'CommonRegexes'}{'ClientSuppliedFilename'},
    'CommandLine'  => $$phProperties{'CommonRegexes'}{'AnyValueExceptNothing'},
    'Created'      => $$phProperties{'CommonRegexes'}{'DateTime'},
    'Creator'      => $$phProperties{'CommonRegexes'}{'Username'},
  );

  $sSafeJobCount = 0;

  foreach my $sTodoJob (@aTodoJobs)
  {
    my (%hJobProperties, %hLArgs, $sNextJob, $sJobId);
    if ($sTodoJob !~ /($$phProperties{'CommonRegexes'}{'JqdQueueTag'})$/)
    {
      $sTodoJobCount--; # Ignore junk in the queue, but decrement the job count to keep the books square.
      next;
    }
    $sJobId = $1;
    if (!-f $sTodoJob || !-W _)
    {
      $$psError = "The job file ($sTodoJob) does not exist or is not writable";
      last;
    }
    %hLArgs =
    (
      'File'           => $sTodoJob,
      'Properties'     => \%hJobProperties,
      'RequireAllKeys' => 1,
      'Template'       => \%hJobTemplate,
      'VerifyValues'   => 1,
    );
    if (!KvpGetKvps(\%hLArgs))
    {
      $$psError = $hLArgs{'Error'};
      last;
    }
    ($sNextJob = $sTodoJob) =~ s/todo\/($$phProperties{'CommonRegexes'}{'JqdQueueTag'})$/sent\/$1/;
    if (!rename($sTodoJob, $sNextJob))
    {
      $$psError = "Unable to move the job ($sTodoJob) to the next queue state ($!)";
      last;
    }
    push(@aJobList, ($sJobId . "=" . $hJobProperties{'CommandLine'} . "\n"));
    $sSafeJobCount++;
  }

  $$phProperties{'JqdJobList'} = join("", @aJobList);

  ####################################################################
  #
  # Perform a sanity check on the results. If there was junk in the
  # queue, it's possible to get a part/whole of 0/0 here. If that's
  # the case, just report that the queue is empty.
  #
  ####################################################################

  if ($sSafeJobCount == 0 && $sTodoJobCount == 0)
  {
    $$psError = substr($$phProperties{'JqdJobType'}, 0, 1) . "-queue is empty"; # This is put inside ()s by the caller.
  }

  ####################################################################
  #
  # Unlock the queue.
  #
  ####################################################################

  JqdUnlockFile(\%hQueueLockArgs);

  return "$sSafeJobCount/$sTodoJobCount";
}


######################################################################
#
# CreateRunTimeEnvironment
#
######################################################################

sub CreateRunTimeEnvironment
{
  my ($phProperties, $psError) = @_;

  ####################################################################
  #
  # Put input/output streams in binary mode.
  #
  ####################################################################

  foreach my $sHandle (\*STDIN, \*STDOUT, \*STDERR)
  {
    binmode($sHandle);
  }

  ####################################################################
  #
  # Initialize regex variables.
  #
  ####################################################################

  my %hCommonRegexes =
  (
    'AnyValue'               => qq(.*),
    'AnyValueExceptNothing'  => qq(.+),
    'Base64'                 => qq([0-9A-Za-z+\/]+={0,2}),
    'BaseDirectory'          => qq((?:[A-Za-z]:)?/[\\w./-]+),
    'ClientId'               => qq((?:[A-Za-z](?:(?:[0-9A-Za-z]|[_-](?=[^.]))){0,62})(?:[.][A-Za-z](?:(?:[0-9A-Za-z]|[_-](?=[^.]))){0,62}){0,127}),
    'ClientSuppliedFilename' => qq([\\w+.:-]{1,1024}),
    'ConfigSearchOrder'      => qq((?:clients(?::commands)?|commands(?::clients)?)),
    'DateTime'               => qq([0-9]{4}-[0-9]{2}-[0-9]{2} [0-9]{2}:[0-9]{2}:[0-9]{2}),
    'Decimal16Bit'           => qq(\\d{1,5}),  # 65535
    'Decimal32Bit'           => qq(\\d{1,10}), # 4294967295
    'Decimal64Bit'           => qq(\\d{1,20}), # 18446744073709551615
    'FilePath'               => qq((?:[A-Za-z]:)?/[\\w./-]+),
    'FolderList'             => qq([\\w.-]+(?::[\\w.-]+)*),
    'Ip'                     => qq((?:\\d{1,3}\\.){3}\\d{1,3}),
    'JobId'                  => qq([\\w-]{1,64}_\\d{10}_\\d{5}),
    'JqdQueueTag'            => qq([ps]\\d{3,5}_(?:\\d{2}_)?\\d{10}_(?:\\d{6}_)?[\\dA-Fa-f]{5,8}),
    'JqdQueueTagOrNa'        => qq((?:NA|[ps]\\d{3,5}_(?:\\d{2}_)?\\d{10}_(?:\\d{6}_)?[\\dA-Fa-f]{5,8})),
    'YesNo'                  => qq([YyNn]),
    'ProcessId'              => qq(\\d{5}),
    'PutNameFormat'          => qq([\\w%+./:-]+),
    'ServerSuppliedPath'     => qq([\\w+./:\\\\-]+),
    'strftime_m'             => qq(\\d{2}),
    'strftime_d'             => qq(\\d{2}),
    'strftime_H'             => qq(\\d{2}),
    'strftime_M'             => qq(\\d{2}),
    'strftime_S'             => qq(\\d{2}),
    'strftime_s'             => qq(\\d{10}),
    'strftime_u'             => qq(\\d{1}),
    'strftime_w'             => qq(\\d{1}),
    'strftime_Y'             => qq(\\d{4}),
    'FileSuffix'             => qq([\\w.-]{1,32}),
    'SystemVersion'          => qq([\\w ()+,./:-]{1,64}),
    'Username'               => qq([\\w]{1,32}),
    'WebJobVersion'          => qq(webjob[\\w (),.-]{1,64}),
  );
  $$phProperties{'CommonRegexes'} = { %hCommonRegexes };

  my %hCustomRegexes =
  (
    'Version'   => qq(VERSION=($$phProperties{'CommonRegexes'}{'WebJobVersion'})),
    'System'    => qq(&SYSTEM=($$phProperties{'CommonRegexes'}{'SystemVersion'})),
    'ClientId'  => qq(&CLIENTID=($$phProperties{'CommonRegexes'}{'ClientId'})),
    'Filename'  => qq(&FILENAME=($$phProperties{'CommonRegexes'}{'ClientSuppliedFilename'})),
    'JqdClientId'  => qq(ClientId=($$phProperties{'CommonRegexes'}{'ClientId'})),
    'JqdJobCount'  => qq(&JobCount=($$phProperties{'CommonRegexes'}{'Decimal16Bit'})),
    'JqdJobType'   => qq(&JobType=(serial|parallel)),
    'JqdQueueName' => qq(&QueueName=((?:$$phProperties{'CommonRegexes'}{'ClientId'}|$$phProperties{'CommonRegexes'}{'ClientSuppliedFilename'}))),
    'QueueTag'  => qq((?:&QUEUETAG=(|$$phProperties{'CommonRegexes'}{'JqdQueueTag'}))?),
    'RunType'   => qq(&RUNTYPE=(linktest|snapshot)),
    'OutLength' => qq(&STDOUT_LENGTH=($$phProperties{'CommonRegexes'}{'Decimal64Bit'})),
    'ErrLength' => qq(&STDERR_LENGTH=($$phProperties{'CommonRegexes'}{'Decimal64Bit'})),
    'EnvLength' => qq(&STDENV_LENGTH=($$phProperties{'CommonRegexes'}{'Decimal64Bit'})),
  );
  $$phProperties{'CustomRegexes'} = { %hCustomRegexes };

  $$phProperties{'CustomRegexes'}{'GetQuery'} =
    $$phProperties{'CustomRegexes'}{'Version'} .
    $$phProperties{'CustomRegexes'}{'System'} .
    $$phProperties{'CustomRegexes'}{'ClientId'} .
    $$phProperties{'CustomRegexes'}{'Filename'} .
    $$phProperties{'CustomRegexes'}{'QueueTag'}
    ;

  $$phProperties{'CustomRegexes'}{'JobQuery'} =
    $$phProperties{'CustomRegexes'}{'JqdClientId'} .
    $$phProperties{'CustomRegexes'}{'JqdQueueName'} .
    $$phProperties{'CustomRegexes'}{'JqdJobType'} .
    $$phProperties{'CustomRegexes'}{'JqdJobCount'}
    ;

  $$phProperties{'CustomRegexes'}{'PutQuery'} =
    $$phProperties{'CustomRegexes'}{'Version'} .
    $$phProperties{'CustomRegexes'}{'System'} .
    $$phProperties{'CustomRegexes'}{'ClientId'} .
    $$phProperties{'CustomRegexes'}{'Filename'} .
    $$phProperties{'CustomRegexes'}{'RunType'} .
    $$phProperties{'CustomRegexes'}{'OutLength'} .
    $$phProperties{'CustomRegexes'}{'ErrLength'} .
    $$phProperties{'CustomRegexes'}{'EnvLength'} .
    $$phProperties{'CustomRegexes'}{'QueueTag'}
    ;

  ####################################################################
  #
  # Initialize environment-specific variables. Pull in SSL-related
  # variables, but only if HTTPS is defined and on.
  #
  ####################################################################

  $$phProperties{'ContentLength'}  = $ENV{'CONTENT_LENGTH'};
  $$phProperties{'Https'}          = $ENV{'HTTPS'};
  $$phProperties{'QueryString'}    = $ENV{'QUERY_STRING'};
  $$phProperties{'RemoteAddress'}  = $ENV{'REMOTE_ADDR'};
  $$phProperties{'RemoteUser'}     = $ENV{'REMOTE_USER'};
  $$phProperties{'RequestMethod'}  = $ENV{'REQUEST_METHOD'} || "";
  $$phProperties{'ServerSoftware'} = $ENV{'SERVER_SOFTWARE'};
  $$phProperties{'PropertiesFile'} = $ENV{'WEBJOB_PROPERTIES_FILE'};

  if (defined($hProperties{'Https'}) && $hProperties{'Https'} =~ /^[Oo][Nn]$/)
  {
    $$phProperties{'SslClientSDnCn'} = $ENV{'SSL_CLIENT_S_DN_CN'};
  }

  ####################################################################
  #
  # Initialize platform-specific variables.
  #
  ####################################################################

  if ($^O =~ /MSWin32/i)
  {
    $$phProperties{'OSClass'} = "WINDOWS";
    $$phProperties{'Newline'} = "\r\n";
    $$phProperties{'NullDevice'} = "nul";
  }
  else
  {
    $$phProperties{'OSClass'} = "UNIX";
    $$phProperties{'Newline'} = "\n";
    $$phProperties{'NullDevice'} = "/dev/null";
    umask(022);
  }

  ####################################################################
  #
  # Initialize site-specific variables. Note that the properties
  # listed in the custom template are a subset of those in the
  # global template. All values in the custom template may may be
  # overridden through the use of client- and/or command-specific
  # config files.
  #
  ####################################################################

  my (%hCustomConfigTemplate, %hGlobalConfigTemplate, $sLocalError);

  %hGlobalConfigTemplate = # This is the set of site-wide properties.
  (
    'BaseDirectory'          => $$phProperties{'CommonRegexes'}{'BaseDirectory'},
    'CapContentLength'       => $$phProperties{'CommonRegexes'}{'YesNo'},
    'ConfigDirectory'        => $$phProperties{'CommonRegexes'}{'BaseDirectory'},
    'ConfigSearchOrder'      => $$phProperties{'CommonRegexes'}{'ConfigSearchOrder'},
    'DsvMaxSignatureLength'  => $$phProperties{'CommonRegexes'}{'Decimal32Bit'},
    'DsvRequireSignatures'   => $$phProperties{'CommonRegexes'}{'YesNo'},
    'DsvSignatureSuffix'     => $$phProperties{'CommonRegexes'}{'FileSuffix'},
    'DynamicDirectory'       => $$phProperties{'CommonRegexes'}{'BaseDirectory'},
    'EnableGetService'       => $$phProperties{'CommonRegexes'}{'YesNo'},
    'EnableHostAccessList'   => $$phProperties{'CommonRegexes'}{'YesNo'},
    'EnableConfigOverrides'  => $$phProperties{'CommonRegexes'}{'YesNo'},
    'EnableJobIds'           => $$phProperties{'CommonRegexes'}{'YesNo'},
    'EnableJobQueues'        => $$phProperties{'CommonRegexes'}{'YesNo'},
    'EnableLogging'          => $$phProperties{'CommonRegexes'}{'YesNo'},
    'EnablePutService'       => $$phProperties{'CommonRegexes'}{'YesNo'},
    'FolderList'             => $$phProperties{'CommonRegexes'}{'FolderList'},
    'GetHookCommandLine'     => $$phProperties{'CommonRegexes'}{'AnyValue'},
    'GetHookLogDivertedOutput' => $$phProperties{'CommonRegexes'}{'YesNo'},
    'GetHookEnable'          => $$phProperties{'CommonRegexes'}{'YesNo'},
    'GetHookStatus'          => $$phProperties{'CommonRegexes'}{'Decimal32Bit'},
    'GetTriggerCommandLine'  => $$phProperties{'CommonRegexes'}{'AnyValue'},
    'GetTriggerEnable'       => $$phProperties{'CommonRegexes'}{'YesNo'},
    'IncomingDirectory'      => $$phProperties{'CommonRegexes'}{'BaseDirectory'},
    'JobQueueActive'         => $$phProperties{'CommonRegexes'}{'YesNo'},
    'JobQueueDirectory'      => $$phProperties{'CommonRegexes'}{'BaseDirectory'},
    'JobQueuePqActiveLimit'  => $$phProperties{'CommonRegexes'}{'Decimal16Bit'},
    'JobQueuePqAnswerLimit'  => $$phProperties{'CommonRegexes'}{'Decimal16Bit'},
    'JobQueueSqActiveLimit'  => $$phProperties{'CommonRegexes'}{'Decimal16Bit'},
    'JobQueueSqAnswerLimit'  => $$phProperties{'CommonRegexes'}{'Decimal16Bit'},
    'LogfilesDirectory'      => $$phProperties{'CommonRegexes'}{'BaseDirectory'},
    'MaxContentLength'       => $$phProperties{'CommonRegexes'}{'Decimal64Bit'},
    'OverwriteExistingFiles' => $$phProperties{'CommonRegexes'}{'YesNo'},
    'ProfilesDirectory'      => $$phProperties{'CommonRegexes'}{'BaseDirectory'},
    'PutNameFormat'          => $$phProperties{'CommonRegexes'}{'PutNameFormat'},
    'PutTriggerCommandLine'  => $$phProperties{'CommonRegexes'}{'AnyValue'},
    'PutTriggerEnable'       => $$phProperties{'CommonRegexes'}{'YesNo'},
    'RequireMatch'           => $$phProperties{'CommonRegexes'}{'YesNo'},
    'RequireUser'            => $$phProperties{'CommonRegexes'}{'YesNo'},
    'ServerId'               => $$phProperties{'CommonRegexes'}{'ClientId'},
    'SpoolDirectory'         => $$phProperties{'CommonRegexes'}{'BaseDirectory'},
    'SslRequireSsl'          => $$phProperties{'CommonRegexes'}{'YesNo'},
    'SslRequireCn'           => $$phProperties{'CommonRegexes'}{'YesNo'},
    'SslRequireMatch'        => $$phProperties{'CommonRegexes'}{'YesNo'},
    'UseGmt'                 => $$phProperties{'CommonRegexes'}{'YesNo'},
  );
  $$phProperties{'GlobalConfigTemplate'} = { %hGlobalConfigTemplate };

  %hCustomConfigTemplate = # This is the subset of site-wide properties that can be overridden.
  (
    'CapContentLength'       => $$phProperties{'CommonRegexes'}{'YesNo'},
    'EnableGetService'       => $$phProperties{'CommonRegexes'}{'YesNo'},
    'EnablePutService'       => $$phProperties{'CommonRegexes'}{'YesNo'},
    'FolderList'             => $$phProperties{'CommonRegexes'}{'FolderList'},
    'GetHookCommandLine'     => $$phProperties{'CommonRegexes'}{'AnyValue'},
    'GetHookLogDivertedOutput' => $$phProperties{'CommonRegexes'}{'YesNo'},
    'GetHookEnable'          => $$phProperties{'CommonRegexes'}{'YesNo'},
    'GetHookStatus'          => $$phProperties{'CommonRegexes'}{'Decimal32Bit'},
    'GetTriggerCommandLine'  => $$phProperties{'CommonRegexes'}{'AnyValue'},
    'GetTriggerEnable'       => $$phProperties{'CommonRegexes'}{'YesNo'},
    'JobQueueActive'         => $$phProperties{'CommonRegexes'}{'YesNo'},
    'JobQueuePqActiveLimit'  => $$phProperties{'CommonRegexes'}{'Decimal16Bit'},
    'JobQueuePqAnswerLimit'  => $$phProperties{'CommonRegexes'}{'Decimal16Bit'},
    'JobQueueSqActiveLimit'  => $$phProperties{'CommonRegexes'}{'Decimal16Bit'},
    'JobQueueSqAnswerLimit'  => $$phProperties{'CommonRegexes'}{'Decimal16Bit'},
    'MaxContentLength'       => $$phProperties{'CommonRegexes'}{'Decimal64Bit'},
    'OverwriteExistingFiles' => $$phProperties{'CommonRegexes'}{'YesNo'},
    'PutNameFormat'          => $$phProperties{'CommonRegexes'}{'PutNameFormat'},
    'PutTriggerCommandLine'  => $$phProperties{'CommonRegexes'}{'AnyValue'},
    'PutTriggerEnable'       => $$phProperties{'CommonRegexes'}{'YesNo'},
    'RequireMatch'           => $$phProperties{'CommonRegexes'}{'YesNo'},
    'RequireUser'            => $$phProperties{'CommonRegexes'}{'YesNo'},
    'SslRequireCn'           => $$phProperties{'CommonRegexes'}{'YesNo'},
    'SslRequireMatch'        => $$phProperties{'CommonRegexes'}{'YesNo'},
  );
  $$phProperties{'CustomConfigTemplate'} = { %hCustomConfigTemplate };

  GetGlobalConfigProperties($phProperties, \%hGlobalConfigTemplate, \$sLocalError);

  ####################################################################
  #
  # Initialize derived variables. If key directories weren't defined
  # in the config file, define them relative to BaseDirectory.
  #
  ####################################################################

  foreach my $sName ("Config", "Dynamic", "Incoming", "Logfiles", "Profiles", "Spool")
  {
    my $sKey = $sName . "Directory";
    if (!exists($$phProperties{$sKey}) || !defined($$phProperties{$sKey}) || length($$phProperties{$sKey}) < 1)
    {
      $$phProperties{$sKey} = $$phProperties{'BaseDirectory'} . "/" . lc($sName);
    }
  }

  if (!exists($$phProperties{'JobQueueDirectory'}) || !defined($$phProperties{'JobQueueDirectory'}) || length($$phProperties{'JobQueueDirectory'}) < 1)
  {
    $$phProperties{'JobQueueDirectory'} = $$phProperties{'SpoolDirectory'} . "/" . "jqd";
  }

  $$phProperties{'NphConfigDirectory'} = $$phProperties{'ConfigDirectory'} . "/" . "nph-webjob";
  $$phProperties{'LogFile'} = $$phProperties{'LogfilesDirectory'} . "/nph-webjob.log";
  $$phProperties{'HookErrFile'} = $$phProperties{'LogfilesDirectory'} . "/nph-webjob-hook.err";
  $$phProperties{'HookLogFile'} = $$phProperties{'LogfilesDirectory'} . "/nph-webjob-hook.log";
  $$phProperties{'HookOutFile'} = $$phProperties{'LogfilesDirectory'} . "/nph-webjob-hook.out";
  $$phProperties{'TriggerLogFile'} = $$phProperties{'LogfilesDirectory'} . "/nph-webjob-trigger.log";

  ####################################################################
  #
  # Conditionally initialize the host access list.
  #
  ####################################################################

  if (Yes($$phProperties{'EnableHostAccessList'}))
  {
    if (!defined(SetupHostAccessList($phProperties, \$sLocalError)))
    {
      $$psError = $sLocalError;
      return undef;
    }
  }

  ####################################################################
  #
  # Verify run time environment.
  #
  ####################################################################

  if (!defined(VerifyRunTimeEnvironment($phProperties, \%hGlobalConfigTemplate, \$sLocalError)))
  {
    $$psError = $sLocalError;
    return undef;
  }

  ####################################################################
  #
  # Conditionally initialize and verify the job id.
  #
  ####################################################################

  if (Yes($$phProperties{'EnableJobIds'}))
  {
    if ($$phProperties{'RequestMethod'} eq "GET")
    {
      $$phProperties{'JobId'} = sprintf("%s_%010u_%05d", $$phProperties{'ServerId'}, $$phProperties{'StartTime'}, $$);
    }
    else
    {
      if (exists($ENV{'HTTP_JOB_ID'}))
      {
        if ($ENV{'HTTP_JOB_ID'} =~ /^($$phProperties{'CommonRegexes'}{'JobId'})$/)
        {
          $$phProperties{'JobId'} = $1;
        }
        else
        {
          $$psError = "HTTP_JOB_ID ($ENV{'HTTP_JOB_ID'}) is undefined or invalid";
          return undef;
        }
      }
      else
      {
        $$phProperties{'JobId'} = "NA"; # Not Assigned.
      }
    }
  }
  else
  {
    $$phProperties{'JobId'} = "NR"; # Not Required.
  }
  if (!defined($$phProperties{'JobId'}) || $$phProperties{'JobId'} !~ /^(NA|NR|$$phProperties{'CommonRegexes'}{'JobId'})$/)
  {
    $$psError = "JobId ($$phProperties{'JobId'}) is undefined or invalid";
    return undef;
  }

  ####################################################################
  #
  # Initialize the job queue tag to a default value. This is needed
  # so that hooks and triggers have something to expand even if job
  # queues are disabled.
  #
  ####################################################################

  $$phProperties{'JqdQueueTag'} = "NA"; # Not Assigned.

  1;
}


######################################################################
#
# ExpandConversionString
#
######################################################################

sub ExpandConversionString
{
  my ($sConversionString, $phConversionValues, $psError) = @_;

  ####################################################################
  #
  # Make sure that required inputs are defined.
  #
  ####################################################################

  if (!defined($sConversionString) || !defined($phConversionValues))
  {
    $$psError = "Unable to proceed in ExpandConversionString() due to missing or undefined inputs";
    return undef;
  }

  ####################################################################
  #
  # Expand the provided conversion string. The TokenList must be
  # processed in reverse order (i.e., from longest to shortest).
  # Otherwise, a token such as %pid would be interpreted as the token
  # %p followed by the literal string "id". Once all regular
  # conversions are done, check for and convert any literal '%'s.
  #
  ####################################################################

  my ($sExpandedConversionString, $sTokenList);

  $sTokenList = join('|', reverse(sort(keys(%$phConversionValues))));
  $sExpandedConversionString = $sConversionString;
  $sExpandedConversionString =~ s/%($sTokenList)/$$phConversionValues{$1}/ge;
  $sExpandedConversionString =~ s/%%/%/g;

  return $sExpandedConversionString;
}


######################################################################
#
# ExpandGetHookCommandLine
#
######################################################################

sub ExpandGetHookCommandLine
{
  my ($phProperties, $psError) = @_;

  ####################################################################
  #
  # Make sure that required inputs are defined.
  #
  ####################################################################

  my %hLArgs =
  (
    'Hash' => $phProperties,
    'Keys' =>
    [
      'ClientFilename',
      'ClientId',
      'ClientSystem',
      'DynamicDirectory',
      'CommonRegexes',
      'GetHookCommandLine',
      'JobId',
      'JqdQueueTag',
      'RemoteAddress',
      'ServerId',
      'HookEpoch',
      'UseGmt',
    ],
  );
  if (!defined(VerifyHashKeys(\%hLArgs)))
  {
    $$psError = $hLArgs{'Error'} if (defined($phProperties));
    return undef;
  }

  ####################################################################
  #
  # Create conversion values.
  #
  ####################################################################

  my
  (
    $sSecond,
    $sMinute,
    $sHour,
    $sMonthDay,
    $sMonth,
    $sYear,
    $sWeekDay,
    $sYearDay,
    $sDaylightSavings,
  ) = (Yes($$phProperties{'UseGmt'})) ? gmtime($$phProperties{'HookEpoch'}) : localtime($$phProperties{'HookEpoch'});

  my %hConversionValues =
  (
    'cid' => $$phProperties{'ClientId'},
    'cmd' => $$phProperties{'ClientFilename'},
    'd'   => sprintf("%02d", $sMonthDay),
    'dynamic_cmd' => $$phProperties{'DynamicDirectory'} . "/" . $$phProperties{'ClientId'} . "/" . "commands" . "/" . $$phProperties{'ClientFilename'},
    'dynamic_dir' => $$phProperties{'DynamicDirectory'},
    'H'   => sprintf("%02d", $sHour),
    'ip'  => $$phProperties{'RemoteAddress'},
    'jid' => $$phProperties{'JobId'},
    'M'   => sprintf("%02d", $sMinute),
    'm'   => sprintf("%02d", $sMonth + 1),
    'pid' => sprintf("%05d", $$),
    'jqt' => $$phProperties{'JqdQueueTag'},
    'S'   => sprintf("%02d", $sSecond),
    's'   => sprintf("%010u", $$phProperties{'HookEpoch'}),
    'sid' => $$phProperties{'ServerId'},
    'system_version' => $$phProperties{'ClientSystem'},
    'u'   => sprintf("%d", $sWeekDay + 1),
    'w'   => sprintf("%d", $sWeekDay),
    'Y'   => sprintf("%04d", $sYear + 1900),
  );

  ####################################################################
  #
  # Verify conversion values.
  #
  ####################################################################

  my ($sLocalError);

  my %hConversionChecks =
  (
    'cid' => $$phProperties{'CommonRegexes'}{'ClientId'},
    'cmd' => $$phProperties{'CommonRegexes'}{'ClientSuppliedFilename'},
    'd'   => $$phProperties{'CommonRegexes'}{'strftime_d'},
    'dynamic_cmd' => $$phProperties{'CommonRegexes'}{'FilePath'},
    'dynamic_dir' => $$phProperties{'CommonRegexes'}{'FilePath'},
    'H'   => $$phProperties{'CommonRegexes'}{'strftime_H'},
    'ip'  => $$phProperties{'CommonRegexes'}{'Ip'},
    'jid' => $$phProperties{'CommonRegexes'}{'JobId'},
    'jqt' => $$phProperties{'CommonRegexes'}{'JqdQueueTagOrNa'},
    'M'   => $$phProperties{'CommonRegexes'}{'strftime_M'},
    'm'   => $$phProperties{'CommonRegexes'}{'strftime_m'},
    'pid' => $$phProperties{'CommonRegexes'}{'ProcessId'},
    'S'   => $$phProperties{'CommonRegexes'}{'strftime_S'},
    's'   => $$phProperties{'CommonRegexes'}{'strftime_s'},
    'sid' => $$phProperties{'CommonRegexes'}{'ClientId'},
    'system_version' => $$phProperties{'CommonRegexes'}{'SystemVersion'},
    'u'   => $$phProperties{'CommonRegexes'}{'strftime_u'},
    'w'   => $$phProperties{'CommonRegexes'}{'strftime_w'},
    'Y'   => $$phProperties{'CommonRegexes'}{'strftime_Y'},
  );

  if (!defined(VerifyConversionValues(\%hConversionValues, \%hConversionChecks, \$sLocalError)))
  {
    $$psError = $sLocalError;
    return undef;
  }

  ####################################################################
  #
  # Expand conversion values.
  #
  ####################################################################

  my $sHookCommandLine = ExpandConversionString($$phProperties{'GetHookCommandLine'}, \%hConversionValues, \$sLocalError);
  if (!defined($sHookCommandLine))
  {
    $$psError = $sLocalError;
    return undef;
  }

  return $sHookCommandLine;
}


######################################################################
#
# ExpandGetTriggerCommandLine
#
######################################################################

sub ExpandGetTriggerCommandLine
{
  my ($phProperties, $psError) = @_;

  ####################################################################
  #
  # Make sure that required inputs are defined.
  #
  ####################################################################

  my %hLArgs =
  (
    'Hash' => $phProperties,
    'Keys' =>
    [
      'ClientFilename',
      'ClientId',
      'CommonRegexes',
      'GetTriggerCommandLine',
      'JobId',
      'JqdQueueTag',
      'RemoteAddress',
      'ServerId',
      'TriggerEpoch',
      'UseGmt',
    ],
  );
  if (!defined(VerifyHashKeys(\%hLArgs)))
  {
    $$psError = $hLArgs{'Error'} if (defined($phProperties));
    return undef;
  }

  ####################################################################
  #
  # Create conversion values.
  #
  ####################################################################

  my
  (
    $sSecond,
    $sMinute,
    $sHour,
    $sMonthDay,
    $sMonth,
    $sYear,
    $sWeekDay,
    $sYearDay,
    $sDaylightSavings,
  ) = (Yes($$phProperties{'UseGmt'})) ? gmtime($$phProperties{'TriggerEpoch'}) : localtime($$phProperties{'TriggerEpoch'});

  my %hConversionValues =
  (
    'CID' => $$phProperties{'ClientId'}, # This is a legacy token, and it is being phased out.
    'cid' => $$phProperties{'ClientId'},
    'CMD' => $$phProperties{'ClientFilename'}, # This is a legacy token, and it is being phased out.
    'cmd' => $$phProperties{'ClientFilename'},
    'd'   => sprintf("%02d", $sMonthDay),
    'H'   => sprintf("%02d", $sHour),
    'IP'  => $$phProperties{'RemoteAddress'}, # This is a legacy token, and it is being phased out.
    'ip'  => $$phProperties{'RemoteAddress'},
    'jid' => $$phProperties{'JobId'},
    'jqt' => $$phProperties{'JqdQueueTag'},
    'M'   => sprintf("%02d", $sMinute),
    'm'   => sprintf("%02d", $sMonth + 1),
    'PID' => sprintf("%05d", $$), # This is a legacy token, and it is being phased out.
    'pid' => sprintf("%05d", $$),
    'S'   => sprintf("%02d", $sSecond),
    's'   => sprintf("%010u", $$phProperties{'TriggerEpoch'}),
    'SID' => $$phProperties{'ServerId'}, # This is a legacy token, and it is being phased out.
    'sid' => $$phProperties{'ServerId'},
    'u'   => sprintf("%d", $sWeekDay + 1),
    'w'   => sprintf("%d", $sWeekDay),
    'Y'   => sprintf("%04d", $sYear + 1900),
  );

  ####################################################################
  #
  # Verify conversion values.
  #
  ####################################################################

  my ($sLocalError);

  my %hConversionChecks =
  (
    'CID' => $$phProperties{'CommonRegexes'}{'ClientId'}, # This is a legacy token, and it is being phased out.
    'cid' => $$phProperties{'CommonRegexes'}{'ClientId'},
    'CMD' => $$phProperties{'CommonRegexes'}{'ClientSuppliedFilename'}, # This is a legacy token, and it is being phased out.
    'cmd' => $$phProperties{'CommonRegexes'}{'ClientSuppliedFilename'},
    'd'   => $$phProperties{'CommonRegexes'}{'strftime_d'},
    'H'   => $$phProperties{'CommonRegexes'}{'strftime_H'},
    'IP'  => $$phProperties{'CommonRegexes'}{'Ip'}, # This is a legacy token, and it is being phased out.
    'ip'  => $$phProperties{'CommonRegexes'}{'Ip'},
    'jid' => $$phProperties{'CommonRegexes'}{'JobId'},
    'jqt' => $$phProperties{'CommonRegexes'}{'JqdQueueTagOrNa'},
    'M'   => $$phProperties{'CommonRegexes'}{'strftime_M'},
    'm'   => $$phProperties{'CommonRegexes'}{'strftime_m'},
    'PID' => $$phProperties{'CommonRegexes'}{'ProcessId'}, # This is a legacy token, and it is being phased out.
    'pid' => $$phProperties{'CommonRegexes'}{'ProcessId'},
    'S'   => $$phProperties{'CommonRegexes'}{'strftime_S'},
    's'   => $$phProperties{'CommonRegexes'}{'strftime_s'},
    'SID' => $$phProperties{'CommonRegexes'}{'ClientId'}, # This is a legacy token, and it is being phased out.
    'sid' => $$phProperties{'CommonRegexes'}{'ClientId'},
    'u'   => $$phProperties{'CommonRegexes'}{'strftime_u'},
    'w'   => $$phProperties{'CommonRegexes'}{'strftime_w'},
    'Y'   => $$phProperties{'CommonRegexes'}{'strftime_Y'},
  );

  if (!defined(VerifyConversionValues(\%hConversionValues, \%hConversionChecks, \$sLocalError)))
  {
    $$psError = $sLocalError;
    return undef;
  }

  ####################################################################
  #
  # Expand conversion values.
  #
  ####################################################################

  my $sTriggerCommandLine = ExpandConversionString($$phProperties{'GetTriggerCommandLine'}, \%hConversionValues, \$sLocalError);
  if (!defined($sTriggerCommandLine))
  {
    $$psError = $sLocalError;
    return undef;
  }

  return $sTriggerCommandLine;
}


######################################################################
#
# ExpandPutTriggerCommandLine
#
######################################################################

sub ExpandPutTriggerCommandLine
{
  my ($phProperties, $psError) = @_;

  ####################################################################
  #
  # Make sure that required inputs are defined.
  #
  ####################################################################

  my %hLArgs =
  (
    'Hash' => $phProperties,
    'Keys' =>
    [
      'ClientFilename',
      'ClientId',
      'CommonRegexes',
      'EnvFile',
      'ErrFile',
      'LckFile',
      'PutTriggerCommandLine',
      'JobId',
      'JqdQueueTag',
      'OutFile',
      'RdyFile',
      'RemoteAddress',
      'ServerId',
      'TriggerEpoch',
      'UseGmt',
    ],
  );
  if (!defined(VerifyHashKeys(\%hLArgs)))
  {
    $$psError = $hLArgs{'Error'} if (defined($phProperties));
    return undef;
  }

  ####################################################################
  #
  # Create conversion values.
  #
  ####################################################################

  my
  (
    $sSecond,
    $sMinute,
    $sHour,
    $sMonthDay,
    $sMonth,
    $sYear,
    $sWeekDay,
    $sYearDay,
    $sDaylightSavings,
  ) = (Yes($$phProperties{'UseGmt'})) ? gmtime($$phProperties{'TriggerEpoch'}) : localtime($$phProperties{'TriggerEpoch'});

  my %hConversionValues =
  (
    'CID' => $$phProperties{'ClientId'}, # This is a legacy token, and it is being phased out.
    'cid' => $$phProperties{'ClientId'},
    'CMD' => $$phProperties{'ClientFilename'}, # This is a legacy token, and it is being phased out.
    'cmd' => $$phProperties{'ClientFilename'},
    'd'   => sprintf("%02d", $sMonthDay),
    'env' => $$phProperties{'EnvFile'},
    'err' => $$phProperties{'ErrFile'},
    'H'   => sprintf("%02d", $sHour),
    'IP'  => $$phProperties{'RemoteAddress'}, # This is a legacy token, and it is being phased out.
    'ip'  => $$phProperties{'RemoteAddress'},
    'jid' => $$phProperties{'JobId'},
    'jqt' => $$phProperties{'JqdQueueTag'},
    'lck' => $$phProperties{'LckFile'},
    'M'   => sprintf("%02d", $sMinute),
    'm'   => sprintf("%02d", $sMonth + 1),
    'out' => $$phProperties{'OutFile'},
    'PID' => sprintf("%05d", $$), # This is a legacy token, and it is being phased out.
    'pid' => sprintf("%05d", $$),
    'rdy' => $$phProperties{'RdyFile'},
    'S'   => sprintf("%02d", $sSecond),
    's'   => sprintf("%010u", $$phProperties{'TriggerEpoch'}),
    'SID' => $$phProperties{'ServerId'}, # This is a legacy token, and it is being phased out.
    'sid' => $$phProperties{'ServerId'},
    'u'   => sprintf("%d", $sWeekDay + 1),
    'w'   => sprintf("%d", $sWeekDay),
    'Y'   => sprintf("%04d", $sYear + 1900),
  );

  ####################################################################
  #
  # Verify conversion values.
  #
  ####################################################################

  my ($sLocalError);

  my %hConversionChecks =
  (
    'CID' => $$phProperties{'CommonRegexes'}{'ClientId'}, # This is a legacy token, and it is being phased out.
    'cid' => $$phProperties{'CommonRegexes'}{'ClientId'},
    'CMD' => $$phProperties{'CommonRegexes'}{'ClientSuppliedFilename'}, # This is a legacy token, and it is being phased out.
    'cmd' => $$phProperties{'CommonRegexes'}{'ClientSuppliedFilename'},
    'd'   => $$phProperties{'CommonRegexes'}{'strftime_d'},
    'env' => $$phProperties{'CommonRegexes'}{'ServerSuppliedPath'},
    'err' => $$phProperties{'CommonRegexes'}{'ServerSuppliedPath'},
    'H'   => $$phProperties{'CommonRegexes'}{'strftime_H'},
    'IP'  => $$phProperties{'CommonRegexes'}{'Ip'}, # This is a legacy token, and it is being phased out.
    'ip'  => $$phProperties{'CommonRegexes'}{'Ip'},
    'jid' => $$phProperties{'CommonRegexes'}{'JobId'},
    'jqt' => $$phProperties{'CommonRegexes'}{'JqdQueueTagOrNa'},
    'lck' => $$phProperties{'CommonRegexes'}{'ServerSuppliedPath'},
    'M'   => $$phProperties{'CommonRegexes'}{'strftime_M'},
    'm'   => $$phProperties{'CommonRegexes'}{'strftime_m'},
    'out' => $$phProperties{'CommonRegexes'}{'ServerSuppliedPath'},
    'PID' => $$phProperties{'CommonRegexes'}{'ProcessId'}, # This is a legacy token, and it is being phased out.
    'pid' => $$phProperties{'CommonRegexes'}{'ProcessId'},
    'rdy' => $$phProperties{'CommonRegexes'}{'ServerSuppliedPath'},
    'S'   => $$phProperties{'CommonRegexes'}{'strftime_S'},
    's'   => $$phProperties{'CommonRegexes'}{'strftime_s'},
    'SID' => $$phProperties{'CommonRegexes'}{'ClientId'}, # This is a legacy token, and it is being phased out.
    'sid' => $$phProperties{'CommonRegexes'}{'ClientId'},
    'u'   => $$phProperties{'CommonRegexes'}{'strftime_u'},
    'w'   => $$phProperties{'CommonRegexes'}{'strftime_w'},
    'Y'   => $$phProperties{'CommonRegexes'}{'strftime_Y'},
  );

  if (!defined(VerifyConversionValues(\%hConversionValues, \%hConversionChecks, \$sLocalError)))
  {
    $$psError = $sLocalError;
    return undef;
  }

  ####################################################################
  #
  # Expand conversion values.
  #
  ####################################################################

  my $sTriggerCommandLine = ExpandConversionString($$phProperties{'PutTriggerCommandLine'}, \%hConversionValues, \$sLocalError);
  if (!defined($sTriggerCommandLine))
  {
    $$psError = $sLocalError;
    return undef;
  }

  return $sTriggerCommandLine;
}


######################################################################
#
# GetCustomConfigProperties
#
######################################################################

sub GetCustomConfigProperties
{
  my ($phProperties, $phTemplate, $psError) = @_;

  ####################################################################
  #
  # Make sure that required inputs are defined.
  #
  ####################################################################

  foreach my $sInput ($phProperties, $phTemplate, $psError)
  {
    if (!defined($sInput))
    {
      $$psError = "Unable to proceed in GetCustomConfigProperties() due to missing or undefined inputs" if (defined($psError));
      return undef;
    }
  }

  ####################################################################
  #
  # Make sure that required inputs are defined.
  #
  ####################################################################

  my %hLArgs =
  (
    'Hash' => $phProperties,
    'Keys' =>
    [
      'ClientId',
      'ConfigSearchOrder',
      'CommonRegexes',
      'EnableJobQueues',
      ($$phProperties{'RequestMethod'} eq "JOB") ? 'JqdQueueName' : 'ClientFilename',
      'NphConfigDirectory',
      'RequestMethod',
    ],
  );
  if (!defined(VerifyHashKeys(\%hLArgs)))
  {
    $$psError = $hLArgs{'Error'} if (defined($phProperties));
    return undef;
  }

  ####################################################################
  #
  # Search for custom config files using the specified order. As each
  # config file is processed, its values trump those of any that came
  # before -- including any values that came from global config files.
  #
  ####################################################################

  foreach my $sFolder (split(/:/, $$phProperties{'ConfigSearchOrder'}))
  {
    ##################################################################
    #
    # The first config file defined by this loop applies globally.
    # The second applies to a particular client or command, and the
    # third applies to a particular client/command pair. Queues, on
    # the other hand, only use global and per queue config files.
    #
    ##################################################################

    my ($sFile1, $sFile2, $sFile3);

    $sFile1 = $sFile2 = $sFile3 = $$phProperties{'NphConfigDirectory'} . "/" . $sFolder . "/";

    if ($sFolder =~ /^clients$/)
    {
      $sFile1 .= "nph-webjob.cfg";
      $sFile2 .= $$phProperties{'ClientId'} . "/" . "nph-webjob.cfg";
      $sFile3 .= $$phProperties{'ClientId'} . "/" . $$phProperties{'ClientFilename'} . "/" . "nph-webjob.cfg";
    }
    elsif ($sFolder =~ /^commands$/)
    {
      $sFile1 .= "nph-webjob.cfg";
      $sFile2 .= $$phProperties{'ClientFilename'} . "/" . "nph-webjob.cfg";
      $sFile3 .= $$phProperties{'ClientFilename'} . "/" . $$phProperties{'ClientId'} . "/" . "nph-webjob.cfg";
    }
    elsif ($sFolder =~ /^queues$/)
    {
      $sFile1 .= "nph-webjob.cfg";
      $sFile2 .= $$phProperties{'JqdQueueName'} . "/" . "nph-webjob.cfg";
      $sFile3 = undef; # This is not used for queues.
    }
    else
    {
      next; # Ignore invalid directories.
    }

    foreach my $sFile ($sFile1, $sFile2, $sFile3)
    {
      next if (!defined($sFile));

      ################################################################
      #
      # Pull in any externally defined properties according to the
      # specified template. If the template is empty, no properties
      # will be returned.
      #
      ################################################################

      my (%hLArgs, %hCustomProperties);

      %hLArgs =
      (
        'File'           => $sFile,
        'Properties'     => \%hCustomProperties,
        'RequireAllKeys' => 0,
        'Template'       => $phTemplate,
        'VerifyValues'   => 0,
      );
      KvpGetKvps(\%hLArgs);

      ################################################################
      #
      # Validate properties according to the specified template. If
      # the template is empty, nothing happens here. Properties that
      # don't pass muster are deleted.
      #
      ################################################################

      foreach my $sProperty (keys(%$phTemplate))
      {
        my $sValue = $hCustomProperties{$sProperty};
        if (!defined($sValue) || $sValue !~ /$$phTemplate{$sProperty}/)
        {
          delete($hCustomProperties{$sProperty});
        }
      }

      ################################################################
      #
      # Transfer validated properties, if any, to the main hash. This
      # is where the trump action takes place.
      #
      ################################################################

      foreach my $sProperty (keys(%hCustomProperties))
      {
        $$phProperties{$sProperty} = $hCustomProperties{$sProperty};
      }
    }
  }

  1;
}


######################################################################
#
# GetGlobalConfigProperties
#
######################################################################

sub GetGlobalConfigProperties
{
  my ($phProperties, $phTemplate, $psError) = @_;

  ####################################################################
  #
  # BaseDirectory is the epicenter of activity.
  #
  ####################################################################

  $$phProperties{'BaseDirectory'} = "/var/webjob";

  ####################################################################
  #
  # CapContentLength forces the script to abort when ContentLength
  # exceeds MaxContentLength.
  #
  ####################################################################

  $$phProperties{'CapContentLength'} = "N"; # [Y|N]

  ####################################################################
  #
  # ConfigSearchOrder specifies the order in which custom config
  # files are sought out and processed. Custom config files may be
  # used to override a predefined subset of the site-specific
  # properties. The following tree enumerates the locations where
  # global and custom config files may exist.
  #
  # config
  #   |
  #   + nph-webjob
  #       |
  #       - nph-webjob.cfg # applies globally
  #       |
  #       + clients
  #       |   |
  #       |   - nph-webjob.cfg # applies globally
  #       |   |
  #       |   + <client-N>
  #       |       |
  #       |       - nph-webjob.cfg # applies to all commands for <client-N>
  #       |       |
  #       |       + <command-N>
  #       |           |
  #       |           - nph-webjob.cfg # applies only to <client-N>/<command-N>
  #       |
  #       + commands
  #           |
  #           - nph-webjob.cfg # applies globally
  #           |
  #           + <command-N>
  #               |
  #               - nph-webjob.cfg # applies to all clients for <command-N>
  #               |
  #               + <client-N>
  #                   |
  #                   - nph-webjob.cfg # applies only to <command-N>/<client-N>
  #
  # As each config file is processed, its values trump those of any
  # that came before -- including any values that came form global
  # config files. Supported values for this variable are "clients",
  # "commands", "clients:commands", and "commands:clients".
  #
  ####################################################################

  $$phProperties{'ConfigSearchOrder'} = "clients:commands";

  ####################################################################
  #
  # DsvMaxSignatureLength specifies the maximum signature length that
  # the script is willing to allow. If the signature length exceeds
  # this limit (in bytes), the script will abort.
  #
  ####################################################################

  $$phProperties{'DsvMaxSignatureLength'} = 256;

  ####################################################################
  #
  # When active, DsvRequireSignatures forces the script to abort if
  # no signature file is found, or if the signature does not meet
  # basic syntax checks. A signature file must have the same basename
  # as the requested payload, and its suffix must match the value
  # defined by DsvSignatureSuffix.
  #
  ####################################################################

  $$phProperties{'DsvRequireSignatures'} = "N"; # [Y|N]

  ####################################################################
  #
  # DsvSignatureSuffix specifies the suffix assigned to and used by
  # signature files. A signature file must have the same basename as
  # the requested payload, and its suffix must match the value
  # defined by this property.
  #
  ####################################################################

  $$phProperties{'DsvSignatureSuffix'} = ".sig";

  ####################################################################
  #
  # When active, EnableGetService allows the script to accept and
  # process valid GET requests. This control can be overridden, and
  # its intended purpose is to administratively disable access for
  # valid clients (e.g., to force a failover).
  #
  ####################################################################

  $$phProperties{'EnableGetService'} = "Y"; # [Y|N]

  ####################################################################
  #
  # When active, EnableHostAccessList causes the script to consult
  # a host access list to determine whether or not the client will
  # be granted access based on its IP address. Note: This mechanism
  # will fail closed if it's enabled and the access list is missing
  # or does not contain the necessary client ID to IP address
  # mappings.
  #
  ####################################################################

  $$phProperties{'EnableHostAccessList'} = "N"; # [Y|N]

  ####################################################################
  #
  # When active, EnableConfigOverrides causes the script to seek
  # out and process additional client- and/or command-specific config
  # files (see ConfigSearchOrder).
  #
  ####################################################################

  $$phProperties{'EnableConfigOverrides'} = "Y"; # [Y|N]

  ####################################################################
  #
  # When active, EnableJobIds forces the script to generate a job id
  # for each GET request. EnableJobIds will also force the script to
  # abort if a PUT request does not contain a valid job id.
  #
  ####################################################################

  $$phProperties{'EnableJobIds'} = "Y"; # [Y|N]

  ####################################################################
  #
  # When active, EnableJobQueues allows the script to pull jobs from
  # job queues and answer JOB requests.
  #
  ####################################################################

  $$phProperties{'EnableJobQueues'} = "Y"; # [Y|N]

  ####################################################################
  #
  # When active, EnableLogging forces the script to generate a log
  # message for each request. If the designated LogFile can not be
  # opened, the log message will be written to STDERR.
  #
  ####################################################################

  $$phProperties{'EnableLogging'} = "Y"; # [Y|N]

  ####################################################################
  #
  # When active, EnablePutService allows the script to accept and
  # process valid PUT requests. This control can be overridden, and
  # its intended purpose is to administratively disable access for
  # valid clients (e.g., to prevent unwanted uploads).
  #
  ####################################################################

  $$phProperties{'EnablePutService'} = "Y"; # [Y|N]

  ####################################################################
  #
  # FolderList specifies locations where shared programs can be found.
  # If a requested file does not exist in a given client's commands
  # directory, the FolderList is searched according to the order given
  # here. The list delimiter is a colon (e.g., "common:shared").
  #
  ####################################################################

  $$phProperties{'FolderList'} = "common";

  ####################################################################
  #
  # GetHookCommandLine is a string consisting of zero or more
  # conversion specifications optionally interspersed with zero or
  # more plain text characters. The following conversion
  # specifications are supported:
  #
  #   %cid = Client ID as a string
  #   %cmd = Client-requested command as a string
  #   %d   = Day of the month as a decimal number (01-31)
  #   %dynamic_cmd = Full path of the command that is to be created
  #   %dynamic_dir = Full path of the dynamic directory
  #   %H   = Hour as a decimal number (00-23)
  #   %jid = Job ID as a string
  #   %ip  = IP address as a dotted quad string
  #   %M   = Minute as a decimal number (00-59)
  #   %m   = Month as a decimal number (01-12)
  #   %pid = Process ID of server-side CGI script
  #   %S   = Second as a decimal number (00-60)
  #   %s   = Number of seconds since the Epoch
  #   %sid = Server ID as a string
  #   %system_version = System version as given by the client
  #   %u   = Weekday as a decimal number (1-7)
  #   %w   = Weekday as a decimal number (0-6)
  #   %Y   = Year with century as a decimal number
  #
  # For example, the following string:
  #
  #   { if [ "%H"X = "23"X ] ; then cp /<path>/checker %dynamic_cmd ; else { echo ; echo ":" ; } > %dynamic_cmd ; fi ; }
  #
  # will copy the checker command to the filename represented by
  # %dynamic_cmd if the current hour is 23. Otherwise, a script that
  # simpy returns true is created on-the-fly -- effectively turning
  # the job into a NOOP.
  #
  # If the specified command is an empty string, then the hook
  # mechanism is (effectively) disabled, and the condition is logged.
  # However, if the hook is disabled (i.e., GetHookEnable=N), then
  # this control is ignored.
  #
  ####################################################################

  $$phProperties{'GetHookCommandLine'} = "";

  ####################################################################
  #
  # GetHookLogDivertedOutput causes the script to divert any output on
  # stdout/stderr to nph-webjob-hook.{err,out}. Normally, output for
  # these streams is discarded since hook commands should be written
  # to keep them clean (e.g., by logging to a file).
  #
  # If a given hook command is not working properly, you can use this
  # option to capture any output that was written to stdout/stderr.
  # That, in turn, may help you debug the problem.
  #
  # If logging is disabled (i.e., EnableLogging=N), this control is
  # ignored.
  #
  ####################################################################

  $$phProperties{'GetHookLogDivertedOutput'} = "N"; # [Y|N]

  ####################################################################
  #
  # When active, GetHookEnable causes the script to execute the
  # command line specified by GetHookCommandLine. The behavior of the
  # hook mechanism is to launch a subprocess and wait for it to
  # finish. The purpose of the subprocess is to create dynamic
  # content that will be delivered to the client. The hook mechanism
  # is highly configurable -- config file overrides are fully
  # supported, multiple conversion tokens are available, and the user
  # determines what, if any, commands are executed when the hook is
  # activated. Currently, hooks are only activated if they are
  # enabled and a hook command has been defined.
  #
  ####################################################################

  $$phProperties{'GetHookEnable'} = "N"; # [Y|N]

  ####################################################################
  #
  # GetHookStatus determines the meaning of success for the hook. If
  # the specified value does not match the exit status as returned by
  # system(), the hook is considered a failure and the entire job is
  # aborted. Otherwise, processing continues as normal.
  #
  ####################################################################

  $$phProperties{'GetHookStatus'} = 0;

  ####################################################################
  #
  # GetTriggerCommandLine is a string consisting of zero or more
  # conversion specifications optionally interspersed with zero or
  # more plain text characters. The following conversion
  # specifications are supported:
  #
  #   %CID = Alias for %cid (legacy token)
  #   %cid = Client ID as a string
  #   %CMD = Alias for %cmd (legacy token)
  #   %cmd = Client-requested command as a string
  #   %d   = Day of the month as a decimal number (01-31)
  #   %H   = Hour as a decimal number (00-23)
  #   %jid = Job ID as a string
  #   %IP  = Alias for %ip (legacy token)
  #   %ip  = IP address as a dotted quad string
  #   %M   = Minute as a decimal number (00-59)
  #   %m   = Month as a decimal number (01-12)
  #   %PID = Alias for %pid (legacy token)
  #   %pid = Process ID of server-side CGI script
  #   %S   = Second as a decimal number (00-60)
  #   %s   = Number of seconds since the Epoch
  #   %SID = Alias for %sid (legacy token)
  #   %sid = Server ID as a string
  #   %u   = Weekday as a decimal number (1-7)
  #   %w   = Weekday as a decimal number (0-6)
  #   %Y   = Year with century as a decimal number
  #
  # For example, the following string:
  #
  #   echo "%Y-%m-%d %H:%M:%S GET %jid %cid" >> /var/log/%cid.jids
  #
  # will append the current date, time, request method, job ID, and
  # client ID to a client-specific file in /var/log.
  #
  # If the specified command is an empty string, then the trigger
  # mechanism is (effectively) disabled, and the condition is logged.
  # However, if the trigger is disabled (i.e., GetTriggerEnable=N),
  # then this control is ignored.
  #
  # Note: Triggers are not currently supported on Windows platforms.
  #
  ####################################################################

  $$phProperties{'GetTriggerCommandLine'} = "";

  ####################################################################
  #
  # When active, GetTriggerEnable causes the script to execute the
  # command line specified by GetTriggerCommandLine. The behavior of
  # the trigger mechanism is to launch a subprocess and continue with
  # the main line of execution. In particular, the script will not
  # block or wait for the subprocess to finish, nor will it attempt
  # check the status or cleanup after the subprocess. The trigger
  # mechanism is highly configurable -- config file overrides are
  # fully supported, multiple conversion tokens are available, and the
  # user determines what, if any, commands are executed when the
  # trigger is pulled. Currently, triggers are only pulled if they are
  # enabled, a trigger command has been defined, and the HTTP status
  # code is 200.
  #
  # Note: Triggers are not currently supported on Windows platforms.
  #
  ####################################################################

  $$phProperties{'GetTriggerEnable'} = "N"; # [Y|N]

  ####################################################################
  #
  # When active, JobQueueActive causes the script to pull jobs from
  # the specified job queue.
  #
  ####################################################################

  $$phProperties{'JobQueueActive'} = "Y"; # [Y|N]

  ####################################################################
  #
  # JobQueuePqActiveLimit specifies the maximum number of parallel
  # jobs that may be in an active state (i.e., either 'sent' or
  # 'open'). If JobQueueActive is disabled, this control is ignored.
  # A value of zero means there is no limit.
  #
  ####################################################################

  $$phProperties{'JobQueuePqActiveLimit'} = 0;

  ####################################################################
  #
  # JobQueuePqAnswerLimit specifies the maximum number of parallel
  # jobs that will be returned to the client. If JobQueueActive is
  # disabled, this control is ignored. A value of zero means there
  # is no limit.
  #
  ####################################################################

  $$phProperties{'JobQueuePqAnswerLimit'} = 0;

  ####################################################################
  #
  # JobQueueSqActiveLimit specifies the maximum number of serial
  # jobs that may be in an active state (i.e., either 'sent' or
  # 'open'). If JobQueueActive is disabled, this control is ignored.
  # A value of zero means there is no limit.
  #
  ####################################################################

  $$phProperties{'JobQueueSqActiveLimit'} = 1;

  ####################################################################
  #
  # JobQueueSqAnswerLimit specifies the maximum number of serial
  # jobs that will be returned to the client. If JobQueueActive is
  # disabled, this control is ignored. A value of zero means there
  # is no limit.
  #
  ####################################################################

  $$phProperties{'JobQueueSqAnswerLimit'} = 0;

  ####################################################################
  #
  # MaxContentLength specifies the largest upload in bytes the script
  # will accept. If CapContentLength is disabled, this control has no
  # effect.
  #
  ####################################################################

  $$phProperties{'MaxContentLength'} = 100000000; # 100 MB

  ####################################################################
  #
  # When active, OverwriteExistingFiles forces the script to unlink
  # existing files prior to writing the uploaded data. The default
  # PutNameFormat used by this script attempts to prevent filename
  # collisions. However, that behavior is user-defined, and in some
  # cases, it may be desirable to specify a PutNameFormat that is
  # guaranteed to create collisions. In those situations, this
  # control must be enabled to produce the desired outcome (i.e.,
  # allow existing files with the same name to be overwritten).
  #
  ####################################################################

  $$phProperties{'OverwriteExistingFiles'} = "N"; # [Y|N]

  ####################################################################
  #
  # PutNameFormat controls how files are named/saved in the incoming
  # directory. In other words, it controls the directory's layout.
  # Basically, PutNameFormat is a format string consisting of zero or
  # more conversion specifications optionally interspersed with zero
  # or more plain text characters. The following conversion
  # specifications are supported:
  #
  #   %CID = Alias for %cid (legacy token)
  #   %cid = Client ID as a string
  #   %CMD = Alias for %cmd (legacy token)
  #   %cmd = Client-requested command as a string
  #   %d   = Day of the month as a decimal number (01-31)
  #   %H   = Hour as a decimal number (00-23)
  #   %IP  = Alias for %ip (legacy token)
  #   %ip  = IP address as a dotted quad string
  #   %M   = Minute as a decimal number (00-59)
  #   %m   = Month as a decimal number (01-12)
  #   %PID = Alias for %pid (legacy token)
  #   %pid = Process ID of server-side CGI script
  #   %S   = Second as a decimal number (00-60)
  #   %s   = Number of seconds since the Epoch
  #   %SID = Alias for %sid (legacy token)
  #   %sid = Server ID as a string
  #   %u   = Weekday as a decimal number (1-7)
  #   %w   = Weekday as a decimal number (0-6)
  #   %Y   = Year with century as a decimal number
  #
  # For example, the following format string:
  #
  #   "%cmd/%ip_%Y-%m-%d_%H.%M.%S"
  #
  # will cause uploaded files to be stored in sub-directories that
  # correspond to the name of the command executed, and each output
  # filename will consist of an IP address, date, and time.
  #
  # The added flexibility provided by this scheme means that it is
  # possible to create format strings that are problematic. Consider
  # the following string:
  #
  #   "%cid/%cmd"
  #
  # While this is a legal format string, it is likely to cause name
  # collisions (e.g., the same client runs the same command two or
  # more times). Therefore, it is important to create format strings
  # that contain enough job specific information to distinguish one
  # set of uploaded files from another.
  #
  ####################################################################

  $$phProperties{'PutNameFormat'} = "%cid_%Y%m%d%H%M%S_%pid_%cmd";

  ####################################################################
  #
  # PutTriggerCommandLine is a string consisting of zero or more
  # conversion specifications optionally interspersed with zero or
  # more plain text characters. The following conversion
  # specifications are supported:
  #
  #   %CID = Alias for %cid (legacy token)
  #   %cid = Client ID as a string
  #   %CMD = Alias for %cmd (legacy token)
  #   %cmd = Client-requested command as a string
  #   %d   = Day of the month as a decimal number (01-31)
  #   %env = Full path to .env file as a string
  #   %err = Full path to .err file as a string
  #   %H   = Hour as a decimal number (00-23)
  #   %jid = Job ID as a string
  #   %IP  = Alias for %ip (legacy token)
  #   %ip  = IP address as a dotted quad string
  #   %lck = Full path to .lck file as a string
  #   %M   = Minute as a decimal number (00-59)
  #   %m   = Month as a decimal number (01-12)
  #   %out = Full path to .out file as a string
  #   %PID = Alias for %pid (legacy token)
  #   %pid = Process ID of server-side CGI script
  #   %rdy = Full path to .rdy file as a string
  #   %S   = Second as a decimal number (00-60)
  #   %s   = Number of seconds since the Epoch
  #   %SID = Alias for %sid (legacy token)
  #   %sid = Server ID as a string
  #   %u   = Weekday as a decimal number (1-7)
  #   %w   = Weekday as a decimal number (0-6)
  #   %Y   = Year with century as a decimal number
  #
  # For example, the following string:
  #
  #   echo "%Y-%m-%d %H:%M:%S PUT %jid %cid" >> /var/log/%cid.jids
  #
  # will append the current date, time, request method, job ID, and
  # client ID to a client-specific file in /var/log.
  #
  # If the specified command is an empty string, then the trigger
  # mechanism is (effectively) disabled, and the condition is logged.
  # However, if the trigger is disabled (i.e., PutTriggerEnable=N),
  # then this control is ignored.
  #
  # Note: Triggers are not currently supported on Windows platforms.
  #
  ####################################################################

  $$phProperties{'PutTriggerCommandLine'} = "";

  ####################################################################
  #
  # When active, PutTriggerEnable causes the script to execute the
  # command line specified by PutTriggerCommandLine. The behavior of
  # the trigger mechanism is to launch a subprocess and continue with
  # the main line of execution. In particular, the script will not
  # block or wait for the subprocess to finish, nor will it attempt
  # check the status or cleanup after the subprocess. The trigger
  # mechanism is highly configurable -- config file overrides are
  # fully supported, multiple conversion tokens are available, and the
  # user determines what, if any, commands are executed when the
  # trigger is pulled. Currently, triggers are only pulled if they are
  # enabled, a trigger command has been defined, and the HTTP status
  # code is 200.
  #
  # Note: Triggers are not currently supported on Windows platforms.
  #
  ####################################################################

  $$phProperties{'PutTriggerEnable'} = "N"; # [Y|N]

  ####################################################################
  #
  # RequireMatch forces the script to abort unless ClientId matches
  # RemoteUser. When this value is disabled, any authenticated user
  # will be allowed to issue requests for a given client. Disabling
  # RequireUser implicitly disables RequireMatch.
  #
  ####################################################################

  $$phProperties{'RequireMatch'} = "Y"; # [Y|N]

  ####################################################################
  #
  # RequireUser forces the script to abort unless RemoteUser has been
  # set.
  #
  ####################################################################

  $$phProperties{'RequireUser'} = "Y"; # [Y|N]

  ####################################################################
  #
  # ServerId specifies the identity assigned to the WebJob server.
  #
  ####################################################################

  $$phProperties{'ServerId'} = "server_1";

  ####################################################################
  #
  # SslRequireCn forces the script to abort unless SslClientSDnCn has
  # been set.  If SslRequireSsl is disabled, this and all other SSL
  # controls are ignored.
  #
  ####################################################################

  $$phProperties{'SslRequireCn'} = "N"; # [Y|N]

  ####################################################################
  #
  # SslRequireMatch forces the script to abort if ClientId does not
  # match SslClientSDnCn. When this control is disabled, access will
  # be governed by RequireMatch.  Disabling SslRequireCn implicitly
  # disables SslRequireMatch.  Also, if SslRequireSsl is disabled,
  # this and all other SSL controls are ignored.  The SslRequireMatch
  # check is performed prior to (not instead of) the RequireMatch
  # check.
  #
  ####################################################################

  $$phProperties{'SslRequireMatch'} = "N"; # [Y|N]

  ####################################################################
  #
  # SslRequireSsl forces the script to abort unless the client is
  # speaking HTTPS.  Disabling SslRequireSsl implicitly disables
  # all SSL-related controls.
  #
  ####################################################################

  $$phProperties{'SslRequireSsl'} = "Y"; # [Y|N]

  ####################################################################
  #
  # When active, UseGmt forces the script to convert all time values
  # to GMT. Otherwise, time values are converted to local time.
  #
  ####################################################################

  $$phProperties{'UseGmt'} = "N"; # [Y|N]

  ####################################################################
  #
  # Pull in any externally defined properties. These properties trump
  # internally defined properties.
  #
  ####################################################################

  my (%hLArgs);

  if (!exists($$phProperties{'PropertiesFile'}) || !defined($$phProperties{'PropertiesFile'}))
  {
    $$phProperties{'PropertiesFile'} = $$phProperties{'BaseDirectory'} . "/config/nph-webjob/nph-webjob.cfg";
  }

  %hLArgs =
  (
    'File'           => $$phProperties{'PropertiesFile'},
    'Properties'     => $phProperties,
    'RequireAllKeys' => 0,
    'Template'       => $phTemplate,
    'VerifyValues'   => 0,
  );
  KvpGetKvps(\%hLArgs);

  1;
}


######################################################################
#
# HookExecuteCommandLine
#
######################################################################

sub HookExecuteCommandLine
{
  my ($phProperties, $psError) = @_;

  ####################################################################
  #
  # Make sure that required inputs are defined.
  #
  ####################################################################

  my %hLArgs =
  (
    'Hash' => $phProperties,
    'Keys' =>
    [
      'EnableLogging',
      'ExpandHookCommandLineRoutine',
      'OSClass',
    ],
  );
  if (!defined(VerifyHashKeys(\%hLArgs)))
  {
    $$psError = $hLArgs{'Error'} if (defined($phProperties));
    return undef;
  }

  ####################################################################
  #
  # Expand the hook's command line. If the result is undefined or
  # null, abort.
  #
  ####################################################################

  my ($sLocalError);

  $$phProperties{'HookEpoch'} = time();

  $$phProperties{'HookCommandLine'} = &{$$phProperties{'ExpandHookCommandLineRoutine'}}($phProperties, \$sLocalError);
  if (!defined($$phProperties{'HookCommandLine'}))
  {
    $$psError = $sLocalError;
    return undef;
  }

  if (!length($$phProperties{'HookCommandLine'}))
  {
    $$psError = "Command line is undefined or null";
    return undef;
  }

  ####################################################################
  #
  # Create directories, as necessary, to hold the dynamic content.
  #
  ####################################################################

  my ($sMode, $sPath);

  $sMode = 0755;
  $sPath = $$phProperties{'DynamicDirectory'} . "/" . $$phProperties{'ClientId'};
  if (!-d $sPath)
  {
    if (!mkdir($sPath, $sMode))
    {
      $$psError = "Directory ($sPath) could not be created ($!)";
      return undef;
    }
  }

  $sPath .= "/" . "commands";
  if (!-d $sPath)
  {
    if (!mkdir($sPath, $sMode))
    {
      $$psError = "Directory ($sPath) could not be created ($!)";
      return undef;
    }
  }

  ####################################################################
  #
  # Wipe the slate clean -- be sure to unlink signature files too.
  #
  ####################################################################

  my ($sGetFile, $sSigFile);

  $sGetFile = $sPath . "/" . $$phProperties{'ClientFilename'};
  unlink($sGetFile);

  $sSigFile = $sGetFile . $$phProperties{'DsvSignatureSuffix'};
  unlink($sSigFile);

  ####################################################################
  #
  # Dup and divert stdout/stderr. This must be done to prevent the
  # response from becoming polluted by spurious hook output.
  #
  ####################################################################

  my ($sStdErrFile, $sStdOutFile);

  if (!open(STDERR_DUP, ">&STDERR"))
  {
    $$psError = "Unable to dup stderr ($!)";
    return undef;
  }
  if (!open(STDOUT_DUP, ">&STDOUT"))
  {
    $$psError = "Unable to dup stdout ($!)";
    return undef;
  }

  if (Yes($$phProperties{'EnableLogging'}) && Yes($$phProperties{'GetHookLogDivertedOutput'}))
  {
    $sStdErrFile = $$phProperties{'HookErrFile'};
    $sStdOutFile = $$phProperties{'HookOutFile'};
  }
  else
  {
    $sStdErrFile = $$phProperties{'NullDevice'};
    $sStdOutFile = $$phProperties{'NullDevice'};
  }

  if (!open(STDERR, ">> $sStdErrFile"))
  {
    $$psError = "Unable to divert stderr to $sStdErrFile ($!)";
    return undef;
  }
  if (!open(STDOUT, ">> $sStdOutFile"))
  {
    $$psError = "Unable to divert stdout to $sStdOutFile ($!)";
    return undef;
  }

  ####################################################################
  #
  # Execute the hook.
  #
  ####################################################################

  $$phProperties{'HookPidLabel'} = "parent";
  $$phProperties{'HookPid'} = $$;
  $$phProperties{'HookState'} = "hooked";
  $$phProperties{'HookMessage'} = $$phProperties{'HookCommandLine'};
  if (Yes($$phProperties{'EnableLogging'}))
  {
    HookLogMessage($phProperties);
  }
  my $sHookReturn = system($$phProperties{'HookCommandLine'});
  my $sHookStatus = ($sHookReturn >> 8) & 0xff;
  my $sHookSignal = ($sHookReturn & 0x7f);
  my $sHookDumped = ($sHookReturn & 0x80) ? 1 : 0;
  if ($sHookStatus == 255)
  {
    $$phProperties{'HookState'} = "failed";
    $$phProperties{'HookMessage'} = "Unable to execute hook command ($!)";
  }
  else
  {
    $$phProperties{'HookState'} = "reaped";
    $$phProperties{'HookMessage'} = "status($sHookStatus) signal($sHookSignal) coredump($sHookDumped)";
  }
  if (Yes($$phProperties{'EnableLogging'}))
  {
    HookLogMessage($phProperties);
  }

  ####################################################################
  #
  # Restore stdout/stderr and close the duplicate handles.
  #
  ####################################################################

  if (!open(STDERR, ">&STDERR_DUP"))
  {
    $$psError = "Unable to restore stderr ($!)";
    return undef;
  }
  if (!open(STDOUT, ">&STDOUT_DUP"))
  {
    $$psError = "Unable to restore stdout ($!)";
    return undef;
  }
  close(STDERR_DUP);
  close(STDOUT_DUP);

  ####################################################################
  #
  # Make sure the return status matches the expected value.
  #
  ####################################################################

  if ($sHookStatus != $$phProperties{'HookStatus'})
  {
    $$psError = "Hook status mismatch ($sHookStatus != $$phProperties{'HookStatus'})";
    return undef;
  }

  1;
}


######################################################################
#
# HookLogMessage
#
######################################################################

sub HookLogMessage
{
  my ($phProperties) = @_;

  ####################################################################
  #
  # Marshall log arguments.
  #
  ####################################################################

  my (%hLArgs);

  $hLArgs{'LogEpoch'} = $$phProperties{'HookEpoch'};

  $hLArgs{'LogFields'} =
  [
    'JobId',
    'RequestMethod',
    'ClientId',
    'ClientFilename',
    'HookPidLabel',
    'HookPid',
    'HookState',
    'Message',
  ];

  $hLArgs{'LogValues'} =
  {
    'JobId'               => $$phProperties{'JobId'},
    'RequestMethod'       => $$phProperties{'RequestMethod'},
    'ClientId'            => $$phProperties{'ClientId'},
    'ClientFilename'      => $$phProperties{'ClientFilename'},
    'HookPidLabel'        => $$phProperties{'HookPidLabel'},
    'HookPid'             => $$phProperties{'HookPid'},
    'HookState'           => $$phProperties{'HookState'},
    'Message'             => $$phProperties{'HookMessage'},
  };

  $hLArgs{'LogFile'} = $$phProperties{'HookLogFile'};

  $hLArgs{'Newline'} = $$phProperties{'Newline'};

  $hLArgs{'RevertToStderr'} = 1;

  $hLArgs{'UseGmt'} = (Yes($$phProperties{'UseGmt'})) ? 1 : 0;

  ####################################################################
  #
  # Deliver log message.
  #
  ####################################################################

  LogNf1vMessage(\%hLArgs);

  1;
}


######################################################################
#
# MakePutName
#
######################################################################

sub MakePutName
{
  my ($phProperties, $psError) = @_;

  ####################################################################
  #
  # Make sure that required inputs are defined.
  #
  ####################################################################

  my %hLArgs =
  (
    'Hash' => $phProperties,
    'Keys' =>
    [
      'ClientFilename',
      'ClientId',
      'CommonRegexes',
      'PutNameFormat',
      'RemoteAddress',
      'ServerId',
      'StartTime',
      'UseGmt',
    ],
  );
  if (!defined(VerifyHashKeys(\%hLArgs)))
  {
    $$psError = $hLArgs{'Error'} if (defined($phProperties));
    return undef;
  }

  ####################################################################
  #
  # Create conversion values.
  #
  ####################################################################

  my
  (
    $sSecond,
    $sMinute,
    $sHour,
    $sMonthDay,
    $sMonth,
    $sYear,
    $sWeekDay,
    $sYearDay,
    $sDaylightSavings
  ) = (Yes($$phProperties{'UseGmt'})) ? gmtime($$phProperties{'StartTime'}) : localtime($$phProperties{'StartTime'});

  my %hConversionValues =
  (
    'CID' => $$phProperties{'ClientId'}, # This is a legacy token, and it is being phased out.
    'cid' => $$phProperties{'ClientId'},
    'CMD' => $$phProperties{'ClientFilename'}, # This is a legacy token, and it is being phased out.
    'cmd' => $$phProperties{'ClientFilename'},
    'IP'  => $$phProperties{'RemoteAddress'}, # This is a legacy token, and it is being phased out.
    'ip'  => $$phProperties{'RemoteAddress'},
    'm'   => sprintf("%02d", $sMonth + 1),
    'd'   => sprintf("%02d", $sMonthDay),
    'H'   => sprintf("%02d", $sHour),
    'M'   => sprintf("%02d", $sMinute),
    'PID' => sprintf("%05d", $$), # This is a legacy token, and it is being phased out.
    'pid' => sprintf("%05d", $$),
    'S'   => sprintf("%02d", $sSecond),
    's'   => sprintf("%010u", $$phProperties{'StartTime'}),
    'SID' => $$phProperties{'ServerId'}, # This is a legacy token, and it is being phased out.
    'sid' => $$phProperties{'ServerId'},
    'u'   => sprintf("%d", $sWeekDay + 1),
    'w'   => sprintf("%d", $sWeekDay),
    'Y'   => sprintf("%04d", $sYear + 1900),
  );

  ####################################################################
  #
  # Verify conversion values.
  #
  ####################################################################

  my ($sLocalError);

  my %hConversionChecks =
  (
    'CID' => $$phProperties{'CommonRegexes'}{'ClientId'}, # This is a legacy token, and it is being phased out.
    'cid' => $$phProperties{'CommonRegexes'}{'ClientId'},
    'CMD' => $$phProperties{'CommonRegexes'}{'ClientSuppliedFilename'}, # This is a legacy token, and it is being phased out.
    'cmd' => $$phProperties{'CommonRegexes'}{'ClientSuppliedFilename'},
    'd'   => $$phProperties{'CommonRegexes'}{'strftime_d'},
    'H'   => $$phProperties{'CommonRegexes'}{'strftime_H'},
    'IP'  => $$phProperties{'CommonRegexes'}{'Ip'}, # This is a legacy token, and it is being phased out.
    'ip'  => $$phProperties{'CommonRegexes'}{'Ip'},
    'M'   => $$phProperties{'CommonRegexes'}{'strftime_M'},
    'm'   => $$phProperties{'CommonRegexes'}{'strftime_m'},
    'PID' => $$phProperties{'CommonRegexes'}{'ProcessId'}, # This is a legacy token, and it is being phased out.
    'pid' => $$phProperties{'CommonRegexes'}{'ProcessId'},
    'S'   => $$phProperties{'CommonRegexes'}{'strftime_S'},
    's'   => $$phProperties{'CommonRegexes'}{'strftime_s'},
    'SID' => $$phProperties{'CommonRegexes'}{'ClientId'}, # This is a legacy token, and it is being phased out.
    'sid' => $$phProperties{'CommonRegexes'}{'ClientId'},
    'u'   => $$phProperties{'CommonRegexes'}{'strftime_u'},
    'w'   => $$phProperties{'CommonRegexes'}{'strftime_w'},
    'Y'   => $$phProperties{'CommonRegexes'}{'strftime_Y'},
  );

  if (!defined(VerifyConversionValues(\%hConversionValues, \%hConversionChecks, \$sLocalError)))
  {
    $$psError = $sLocalError;
    return undef;
  }

  ####################################################################
  #
  # Expand conversion values.
  #
  ####################################################################

  my $sPutName = ExpandConversionString($$phProperties{'PutNameFormat'}, \%hConversionValues, \$sLocalError);
  if (!defined($sPutName))
  {
    $$psError = $sLocalError;
    return undef;
  }

  return $sPutName;
}


######################################################################
#
# MakePutTree
#
######################################################################

sub MakePutTree
{
  my ($sIncomingDirectory, $sPutName, $sMode, $sPopCount, $psError) = @_;

  ####################################################################
  #
  # Pop the specified number of elements from PutName. Normally, only
  # the trailing filename is removed (i.e. PopCount = 1).
  #
  ####################################################################

  my (@aElements);

  @aElements = split(/[\/\\]/, $sPutName);

  while (defined($sPopCount) && $sPopCount-- > 0)
  {
    pop(@aElements);
  }

  ####################################################################
  #
  # Create the tree -- one element at a time. If the mkdir() fails,
  # test the directory's existence again -- another process may have
  # beat us to the punch, and that is not a fatal error.
  #
  ####################################################################

  my ($sLocalError, $sPath);

  $sPath = $sIncomingDirectory;

  foreach my $sElement (@aElements)
  {
    $sPath .= "/$sElement";
    foreach my $sTry (0..1)
    {
      if (!-d $sPath && !mkdir($sPath, $sMode))
      {
        $sLocalError = "Directory ($sPath) could not be created ($!)";
      }
      else
      {
        $sLocalError = undef;
        last;
      }
    }
    if (defined($sLocalError))
    {
      $$psError = $sLocalError;
      return undef;
    }
  }

  1;
}


######################################################################
#
# NphLogMessage
#
######################################################################

sub NphLogMessage
{
  my ($phProperties) = @_;

  ####################################################################
  #
  # Marshall log arguments.
  #
  ####################################################################

  my (%hLArgs, $sVariableKey);

  $hLArgs{'LogEpoch'} = time();

  $sVariableKey = ($$phProperties{'RequestMethod'} eq "JOB") ? "JqdQueueName" : "ClientFilename";

  $hLArgs{'LogFields'} =
  [
    'JobId',
    'RemoteUser',
    'RemoteAddress',
    'RequestMethod',
    'ClientId',
    $sVariableKey,
    'ContentLength',
    'ServerContentLength',
    'Duration',
    'ReturnStatus',
    'Message'
  ];

  $hLArgs{'LogValues'} =
  {
    'JobId'               => $$phProperties{'JobId'},
    'RemoteUser'          => $$phProperties{'RemoteUser'},
    'RemoteAddress'       => $$phProperties{'RemoteAddress'},
    'RequestMethod'       => $$phProperties{'RequestMethod'},
    'ClientId'            => $$phProperties{'ClientId'},
    $sVariableKey         => $$phProperties{$sVariableKey},
    'ContentLength'       => $$phProperties{'ContentLength'},
    'ServerContentLength' => $$phProperties{'ServerContentLength'},
    'Duration'            => $hLArgs{'LogEpoch'} - $$phProperties{'StartTime'},
    'ReturnStatus'        => $$phProperties{'ReturnStatus'},
    'Message'             => $$phProperties{'ErrorMessage'},
  };

  $hLArgs{'LogFile'} = $$phProperties{'LogFile'};

  $hLArgs{'Newline'} = $$phProperties{'Newline'};

  $hLArgs{'RevertToStderr'} = 1;

  $hLArgs{'UseGmt'} = (Yes($$phProperties{'UseGmt'})) ? 1 : 0;

  ####################################################################
  #
  # Deliver log message.
  #
  ####################################################################

  LogNf1vMessage(\%hLArgs);

  1;
}


######################################################################
#
# ProcessGetRequest
#
######################################################################

sub ProcessGetRequest
{
  my ($phProperties, $psError) = @_;

  ####################################################################
  #
  # Proceed only if QueryString matches the GetQuery expression.
  #
  ####################################################################

  my $sQueryString = URLDecode($$phProperties{'QueryString'});

  if ($sQueryString =~ /^$$phProperties{'CustomRegexes'}{'GetQuery'}$/)
  {
    $$phProperties{'ClientVersion'}  = $1;
    $$phProperties{'ClientSystem'}   = $2;
    $$phProperties{'ClientId'}       = $3 || "nobody";
    $$phProperties{'ClientFilename'} = $4;
    $$phProperties{'QueueTag'}       = $5 || "";

    ##################################################################
    #
    # Bring in any client- and/or command-specific properties.
    #
    ##################################################################

    my ($sLocalError);

    if (Yes($$phProperties{'EnableConfigOverrides'}))
    {
      GetCustomConfigProperties($phProperties, \%{$$phProperties{'CustomConfigTemplate'}}, \$sLocalError);
    }

    ##################################################################
    #
    # Conditionally do CommonName and client ID checks.
    #
    ##################################################################

    if (Yes($$phProperties{'SslRequireSsl'}))
    {
      if (Yes($$phProperties{'SslRequireCn'}) && (!defined($$phProperties{'SslClientSDnCn'}) || !length($$phProperties{'SslClientSDnCn'})))
      {
        $$psError = "CommonName is undefined or null";
        return 470;
      }

      if (Yes($$phProperties{'SslRequireCn'}) && Yes($$phProperties{'SslRequireMatch'}) && $$phProperties{'SslClientSDnCn'} ne $$phProperties{'ClientId'})
      {
        $$psError = "CommonName ($$phProperties{'SslClientSDnCn'}) does not match client ID ($$phProperties{'ClientId'})";
        return 471;
      }
    }

    ##################################################################
    #
    # Do username and client ID checks.
    #
    ##################################################################

    if (Yes($$phProperties{'RequireUser'}) && (!defined($$phProperties{'RemoteUser'}) || !length($$phProperties{'RemoteUser'})))
    {
      $$psError = "Remote user is undefined or null";
      return 452;
    }

    if (Yes($$phProperties{'RequireUser'}) && Yes($$phProperties{'RequireMatch'}) && $$phProperties{'RemoteUser'} ne $$phProperties{'ClientId'})
    {
      $$psError = "Remote user ($$phProperties{'RemoteUser'}) does not match client ID ($$phProperties{'ClientId'})";
      return 453;
    }

    ##################################################################
    #
    # Do host access list checks.
    #
    ##################################################################

    if (Yes($$phProperties{'EnableHostAccessList'}) && !CheckHostAccessList($$phProperties{'AccessList'}{$$phProperties{'ClientId'}}, $$phProperties{'RemoteAddress'}))
    {
      $$psError = "The ClientId/Address key/value pair ($$phProperties{'ClientId'}=$$phProperties{'RemoteAddress'}/32) is not on the host access list";
      return 403;
    }

    ##################################################################
    #
    # Do content length checks.
    #
    ##################################################################

    if (!defined($$phProperties{'ContentLength'}) || !length($$phProperties{'ContentLength'}))
    {
      $$psError = "Content length is undefined or null";
      return 454;
    }

    if (Yes($$phProperties{'CapContentLength'}) && $$phProperties{'ContentLength'} > $$phProperties{'MaxContentLength'})
    {
      $$psError = "Content length ($$phProperties{'ContentLength'}) exceeds maximum allowed length ($$phProperties{'MaxContentLength'})";
      return 455;
    }

    ##################################################################
    #
    # Confirm that service is currently enabled.
    #
    ##################################################################

    if (!Yes($$phProperties{'EnableGetService'}))
    {
      $$psError = "GET service is disabled";
      return 553;
    }

    ##################################################################
    #
    # Conditionally validate queue properties.
    #
    ##################################################################

    if (Yes($$phProperties{'EnableJobQueues'}) && $$phProperties{'QueueTag'} =~ /^($$phProperties{'CommonRegexes'}{'JqdQueueTag'})$/)
    {
      $$phProperties{'JqdQueueTag'} = $$phProperties{'JqdQueueNid'} = $1; # Assume that $1 is defined and valid.
      $$phProperties{'JqdQueueNid'} =~ s/^[ps]//;
      $$phProperties{'JqdQueueNid'} =~ s/_.*$//;
      my %hLArgs =
      (
        'ClientId'          => $$phProperties{'ClientId'},
        'CommonRegexes'     => $$phProperties{'CommonRegexes'},
        'JobQueueDirectory' => $$phProperties{'JobQueueDirectory'},
        'JqdQueueNid'       => $$phProperties{'JqdQueueNid'},
      );
      $$phProperties{'JqdQueueName'} = JqdGetQueueName(\%hLArgs);
      if (!defined($$phProperties{'JqdQueueName'}))
      {
        if ($hLArgs{'Error'})
        {
          $$psError = $hLArgs{'Error'};
          return 500;
        }
        $$psError = "Queue NID ($$phProperties{'JqdQueueNid'}) is not mapped.";
        return 462;
      }
      $$phProperties{'JqdQueueDirectory'} = $$phProperties{'JobQueueDirectory'} . "/" . $$phProperties{'JqdQueueName'};
      %hLArgs =
      (
        'Directory' => $$phProperties{'JqdQueueDirectory'},
      );
      if (!JqdCheckQueueTree(\%hLArgs))
      {
        $$psError = $hLArgs{'Error'};
        return 461;
      }
      my $sJobFile = $$phProperties{'JqdQueueDirectory'} . "/" . "sent" . "/" . $$phProperties{'JqdQueueTag'};
      if (!-f $sJobFile)
      {
        $$psError = "Queue file ($sJobFile) was not found in the specified queue ($$phProperties{'JqdQueueName'})";
        return 404;
      }
    }

    ##################################################################
    #
    # Locate the requested file and serve it up. Start by searching
    # the client's commands directory. Then, move on to the shared
    # folders. If a GetHook is enabled, insert that path at the head
    # of the folder list since dynamic content trumps everything.
    #
    ##################################################################

    my $sEffectiveFolderList = $$phProperties{'ClientId'} . ":" . $$phProperties{'FolderList'};

    if (Yes($$phProperties{'GetHookEnable'}))
    {
      $$phProperties{'ExpandHookCommandLineRoutine'} = \&ExpandGetHookCommandLine;
      $$phProperties{'HookStatus'} = $$phProperties{'GetHookStatus'};
      if (!HookExecuteCommandLine($phProperties, \$sLocalError))
      {
        $$psError = $sLocalError;
        return 500;
      }
      $sEffectiveFolderList = "../dynamic/" . $$phProperties{'ClientId'} . ":" . $sEffectiveFolderList;
    }

    foreach my $sFolder (split(/:/, $sEffectiveFolderList))
    {
      my $sGetFile = $$phProperties{'ProfilesDirectory'} . "/" . $sFolder . "/" . "commands" . "/" . $$phProperties{'ClientFilename'};
      if (-e $sGetFile)
      {
        my $sSigFile = $sGetFile . $$phProperties{'DsvSignatureSuffix'};
        if (-f $sSigFile && -s _ && open(FH, "< $sSigFile"))
        {
          binmode(FH);
          $$phProperties{'DsvPayloadSignature'} = <FH>; # This file should only contain one line.
          $$phProperties{'DsvPayloadSignature'} =~ s/[\r\n]*$//;
          if
          (
            $$phProperties{'DsvPayloadSignature'} !~ /^$$phProperties{'CommonRegexes'}{'Base64'}$/ ||
            length($$phProperties{'DsvPayloadSignature'}) > $$phProperties{'DsvMaxSignatureLength'}
          )
          {
            $$phProperties{'DsvPayloadSignature'} = undef;
          }
          close(FH);
        }
        else
        {
          $$phProperties{'DsvPayloadSignature'} = undef;
        }
        if (Yes($$phProperties{'DsvRequireSignatures'}) && !defined($$phProperties{'DsvPayloadSignature'}))
        {
          $$psError = "Payload signature ($sSigFile) could not be opened, does not exist, contains invalid data, or is not the correct length";
          return 459;
        }
        if (!open(FH, "< $sGetFile"))
        {
          $$psError = "Requested file ($sGetFile) could not be opened ($!)";
          return 457;
        }
        binmode(FH);
        $$phProperties{'ReturnHandle'} = \*FH;
        if (Yes($$phProperties{'EnableJobQueues'}) && $$phProperties{'JqdQueueTag'} =~ /^$$phProperties{'CommonRegexes'}{'JqdQueueTag'}$/)
        {
          if (!UpdateQueue($phProperties, \$sLocalError))
          {
            $$psError = $sLocalError;
            return 500;
          }
          $$psError = "Success ($$phProperties{'JqdQueueTag'})";
        }
        else
        {
          $$psError = "Success";
        }
        return 200;
      }
    }
    $$psError = "Requested file ($$phProperties{'ClientFilename'}) was not found in effective folder list ($sEffectiveFolderList)";
    return 404;
  }
  elsif ($sQueryString =~ /^$$phProperties{'CustomRegexes'}{'JobQuery'}$/)
  {
    $$phProperties{'ClientId'}     = $1;
    $$phProperties{'JqdQueueName'} = $2;
    $$phProperties{'JqdJobType'}   = $3;
    $$phProperties{'JqdJobCount'}  = $4;

    ##################################################################
    #
    # Convert this GET request to a JOB request, and initialize key
    # properties.
    #
    ##################################################################

    $$phProperties{'RequestMethod'} = "JOB";
    $$phProperties{'JqdJobList'} = ""; # Note: This is referenced in SendResponse().

    ##################################################################
    #
    # Bring in any queue-specific properties.
    #
    ##################################################################

    my ($sLocalError);

    if (Yes($$phProperties{'EnableConfigOverrides'}))
    {
      $$phProperties{'ConfigSearchOrder'} = "queues"; # Confine the search to this one path.
      GetCustomConfigProperties($phProperties, \%{$$phProperties{'CustomConfigTemplate'}}, \$sLocalError);
    }

    ##################################################################
    #
    # Conditionally do CommonName and client ID checks.
    #
    ##################################################################

    if (Yes($$phProperties{'SslRequireSsl'}))
    {
      if (Yes($$phProperties{'SslRequireCn'}) && (!defined($$phProperties{'SslClientSDnCn'}) || !length($$phProperties{'SslClientSDnCn'})))
      {
        $$psError = "CommonName is undefined or null";
        return 470;
      }

      if (Yes($$phProperties{'SslRequireCn'}) && Yes($$phProperties{'SslRequireMatch'}) && $$phProperties{'SslClientSDnCn'} ne $$phProperties{'ClientId'})
      {
        $$psError = "CommonName ($$phProperties{'SslClientSDnCn'}) does not match client ID ($$phProperties{'ClientId'})";
        return 471;
      }
    }

    ##################################################################
    #
    # Do username and client ID checks.
    #
    ##################################################################

    if (Yes($$phProperties{'RequireUser'}) && (!defined($$phProperties{'RemoteUser'}) || !length($$phProperties{'RemoteUser'})))
    {
      $$psError = "Remote user is undefined or null";
      return 452;
    }

    if (Yes($$phProperties{'RequireUser'}) && Yes($$phProperties{'RequireMatch'}) && $$phProperties{'RemoteUser'} ne $$phProperties{'ClientId'})
    {
      $$psError = "Remote user ($$phProperties{'RemoteUser'}) does not match client ID ($$phProperties{'ClientId'})";
      return 453;
    }

    ##################################################################
    #
    # Do host access list checks.
    #
    ##################################################################

    if (Yes($$phProperties{'EnableHostAccessList'}) && !CheckHostAccessList($$phProperties{'AccessList'}{$$phProperties{'ClientId'}}, $$phProperties{'RemoteAddress'}))
    {
      $$psError = "The ClientId/Address key/value pair ($$phProperties{'ClientId'}=$$phProperties{'RemoteAddress'}/32) is not on the host access list";
      return 403;
    }

    ##################################################################
    #
    # Do content length checks.
    #
    ##################################################################

    if (!defined($$phProperties{'ContentLength'}) || !length($$phProperties{'ContentLength'}))
    {
      $$psError = "Content length is undefined or null";
      return 454;
    }

    if (Yes($$phProperties{'CapContentLength'}) && $$phProperties{'ContentLength'} > $$phProperties{'MaxContentLength'})
    {
      $$psError = "Content length ($$phProperties{'ContentLength'}) exceeds maximum allowed length ($$phProperties{'MaxContentLength'})";
      return 455;
    }

    ##################################################################
    #
    # Conditionally retrieve the current job list and serve it up.
    #
    ##################################################################

    if (Yes($$phProperties{'EnableJobQueues'}))
    {
      if (Yes($$phProperties{'JobQueueActive'}))
      {
        $$phProperties{'JqdQueueDirectory'} = $$phProperties{'JobQueueDirectory'} . "/" . $$phProperties{'JqdQueueName'};
        my %hLArgs =
        (
          'Directory' => $$phProperties{'JqdQueueDirectory'},
        );
        if (!JqdCheckQueueTree(\%hLArgs))
        {
          $$psError = $hLArgs{'Error'};
          return 461;
        }
        %hLArgs =
        (
          'ClientId'          => $$phProperties{'ClientId'},
          'CommonRegexes'     => $$phProperties{'CommonRegexes'},
          'JobQueueDirectory' => $$phProperties{'JobQueueDirectory'},
          'JqdQueueName'      => $$phProperties{'JqdQueueName'},
        );
        $$phProperties{'JqdQueueNid'} = JqdGetQueueNid(\%hLArgs);
        if (!defined($$phProperties{'JqdQueueNid'}))
        {
          if ($hLArgs{'Error'})
          {
            $$psError = $hLArgs{'Error'};
            return 500;
          }
          $$psError = "Queue name ($$phProperties{'JqdQueueName'}) is not mapped.";
          return 462;
        }
        my $sJobTally = CreateJobList($phProperties, \$sLocalError);
        if (!defined($sJobTally))
        {
          $$psError = $sLocalError;
          # NOTE: Errors prior to a state transition don't require a freeze.
          return ($sLocalError =~ /frozen/) ? 503 : 500;
        }
        else
        {
          my ($sPart, $sWhole) = ($sJobTally =~ /(\d+)\/(\d+)/);
          if ($sPart == 0 && $sWhole == 0)
          {
            $$psError = "Success ($sLocalError)";
            return 204;
          }
          else
          {
            if ($sPart == $sWhole)
            {
              $$psError = "Success (returned $sJobTally $$phProperties{'JqdJobType'} jobs)";
              return 200;
            }
            else
            {
              $$psError = "Partial Success (returned $sJobTally $$phProperties{'JqdJobType'} jobs) Error='$sLocalError'";
              # NOTE: Errors during a state transition require a freeze.
              my %hLArgs =
              (
                'Directory' => $$phProperties{'JqdQueueDirectory'},
              );
              JqdFreezeQueue(\%hLArgs);
              return 206;
            }
          }
        }
      }
      else
      {
        $$psError = "Queue ($hProperties{'JqdQueueName'}) is not active.";
        return 460;
      }
    }
    else
    {
      $$psError = "Method ($hProperties{'RequestMethod'}) not allowed";
      return 405;
    }
  }
  else
  {
    $$psError = "Invalid query string ($$phProperties{'QueryString'})";
    return 450;
  }
}


######################################################################
#
# ProcessPutRequest
#
######################################################################

sub ProcessPutRequest
{
  my ($phProperties, $psError) = @_;

  ####################################################################
  #
  # Proceed only if QueryString matches the PutQuery expression.
  #
  ####################################################################

  my $sQueryString = URLDecode($$phProperties{'QueryString'});

  if ($sQueryString =~ /^$$phProperties{'CustomRegexes'}{'PutQuery'}$/)
  {
    my ($sEnvLength, $sErrLength, $sOutLength);

    $$phProperties{'ClientVersion'}   = $1;
    $$phProperties{'ClientSystem'}    = $2;
    $$phProperties{'ClientId'}        = $3 || "nobody";
    $$phProperties{'ClientFilename'}  = $4;
    $$phProperties{'ClientRunType'}   = $5;
    $$phProperties{'ClientOutLength'} = $sOutLength = $6;
    $$phProperties{'ClientErrLength'} = $sErrLength = $7;
    $$phProperties{'ClientEnvLength'} = $sEnvLength = $8;
    $$phProperties{'QueueTag'}        = $9 || "";

    ##################################################################
    #
    # Bring in any client- and/or command-specific properties.
    #
    ##################################################################

    my ($sLocalError);

    if (Yes($$phProperties{'EnableConfigOverrides'}))
    {
      GetCustomConfigProperties($phProperties, \%{$$phProperties{'CustomConfigTemplate'}}, \$sLocalError);
    }

    ##################################################################
    #
    # Conditionally do CommonName and client ID checks.
    #
    ##################################################################

    if (Yes($$phProperties{'SslRequireSsl'}))
    {
      if (Yes($$phProperties{'SslRequireCn'}) && (!defined($$phProperties{'SslClientSDnCn'}) || !length($$phProperties{'SslClientSDnCn'})))
      {
        $$psError = "CommonName is undefined or null";
        return 470;
      }

      if (Yes($$phProperties{'SslRequireCn'}) && Yes($$phProperties{'SslRequireMatch'}) && $$phProperties{'SslClientSDnCn'} ne $$phProperties{'ClientId'})
      {
        $$psError = "CommonName ($$phProperties{'SslClientSDnCn'}) does not match client ID ($$phProperties{'ClientId'})";
        return 471;
      }
    }

    ##################################################################
    #
    # Do username and client ID checks.
    #
    ##################################################################

    if (Yes($$phProperties{'RequireUser'}) && (!defined($$phProperties{'RemoteUser'}) || !length($$phProperties{'RemoteUser'})))
    {
      $$psError = "Remote user is undefined or null";
      return 452;
    }

    if (Yes($$phProperties{'RequireUser'}) && Yes($$phProperties{'RequireMatch'}) && $$phProperties{'RemoteUser'} ne $$phProperties{'ClientId'})
    {
      $$psError = "Remote user ($$phProperties{'RemoteUser'}) does not match client ID ($$phProperties{'ClientId'})";
      return 453;
    }

    ##################################################################
    #
    # Do host access list checks.
    #
    ##################################################################

    if (Yes($$phProperties{'EnableHostAccessList'}) && !CheckHostAccessList($$phProperties{'AccessList'}{$$phProperties{'ClientId'}}, $$phProperties{'RemoteAddress'}))
    {
      $$psError = "The ClientId/Address key/value pair ($$phProperties{'ClientId'}=$$phProperties{'RemoteAddress'}/32) is not on the host access list";
      return 403;
    }

    ##################################################################
    #
    # Do content length checks.
    #
    ##################################################################

    if (!defined($$phProperties{'ContentLength'}) || !length($$phProperties{'ContentLength'}))
    {
      $$psError = "Content length is undefined or null";
      return 454;
    }

    if (Yes($$phProperties{'CapContentLength'}) && $$phProperties{'ContentLength'} > $$phProperties{'MaxContentLength'})
    {
      $$psError = "Content length ($$phProperties{'ContentLength'}) exceeds maximum allowed length ($$phProperties{'MaxContentLength'})";
      return 455;
    }

    if ($$phProperties{'ContentLength'} != ($sOutLength + $sErrLength + $sEnvLength))
    {
      $$psError = "Content length ($$phProperties{'ContentLength'}) does not equal sum of individual stream lengths ($sOutLength + $sErrLength + $sEnvLength)";
      return 456;
    }

    ##################################################################
    #
    # Confirm that service is currently enabled.
    #
    ##################################################################

    if (!Yes($$phProperties{'EnablePutService'}))
    {
      $$psError = "PUT service is disabled";
      return 553;
    }

    ##################################################################
    #
    # Conditionally validate queue properties.
    #
    ##################################################################

    if (Yes($$phProperties{'EnableJobQueues'}) && $$phProperties{'QueueTag'} =~ /^($$phProperties{'CommonRegexes'}{'JqdQueueTag'})$/)
    {
      $$phProperties{'JqdQueueTag'} = $$phProperties{'JqdQueueNid'} = $1; # Assume that $1 is defined and valid.
      $$phProperties{'JqdQueueNid'} =~ s/^[ps]//;
      $$phProperties{'JqdQueueNid'} =~ s/_.*$//;
      my %hLArgs =
      (
        'ClientId'          => $$phProperties{'ClientId'},
        'CommonRegexes'     => $$phProperties{'CommonRegexes'},
        'JobQueueDirectory' => $$phProperties{'JobQueueDirectory'},
        'JqdQueueNid'       => $$phProperties{'JqdQueueNid'},
      );
      $$phProperties{'JqdQueueName'} = JqdGetQueueName(\%hLArgs);
      if (!defined($$phProperties{'JqdQueueName'}))
      {
        if ($hLArgs{'Error'})
        {
          $$psError = $hLArgs{'Error'};
          return 500;
        }
        $$psError = "Queue NID ($$phProperties{'JqdQueueNid'}) is not mapped.";
        return 462;
      }
      $$phProperties{'JqdQueueDirectory'} = $$phProperties{'JobQueueDirectory'} . "/" . $$phProperties{'JqdQueueName'};
      %hLArgs =
      (
        'Directory' => $$phProperties{'JqdQueueDirectory'},
      );
      if (!JqdCheckQueueTree(\%hLArgs))
      {
        $$psError = $hLArgs{'Error'};
        return 461;
      }
      my $sJobFile = $$phProperties{'JqdQueueDirectory'} . "/" . "open" . "/" . $$phProperties{'JqdQueueTag'};
      if (!-f $sJobFile)
      {
        $$psError = "Queue file ($sJobFile) was not found in the specified queue ($$phProperties{'JqdQueueName'})";
        return 404;
      }
    }

    ##################################################################
    #
    # If this is a link test, dump the data and return success.
    #
    ##################################################################

    if ($$phProperties{'ClientRunType'} eq "linktest")
    {
      SysReadWrite(\*STDIN, undef, $$phProperties{'ContentLength'}, undef); # Slurp up data to prevent a broken pipe.
      $$psError = "Success";
      $$psError .= " ($$phProperties{'JqdQueueTag'})" if ($$phProperties{'JqdQueueTag'});
      return 251;
    }

    ##################################################################
    #
    # Make output filenames and directories.
    #
    ##################################################################

    my ($sEnvFile, $sErrFile, $sLckFile, $sOutFile, $sPutName, $sRdyFile);

    $sPutName = MakePutName($phProperties, \$sLocalError);
    if (!defined($sPutName))
    {
      $$psError = $sLocalError;
      SysReadWrite(\*STDIN, undef, $$phProperties{'ContentLength'}, undef); # Slurp up data to prevent a broken pipe.
      return 500;
    }
    $sLckFile = $$phProperties{'IncomingDirectory'} . "/" . $sPutName . ".lck";
    $sOutFile = $$phProperties{'IncomingDirectory'} . "/" . $sPutName . ".out";
    $sErrFile = $$phProperties{'IncomingDirectory'} . "/" . $sPutName . ".err";
    $sEnvFile = $$phProperties{'IncomingDirectory'} . "/" . $sPutName . ".env";
    $sRdyFile = $$phProperties{'IncomingDirectory'} . "/" . $sPutName . ".rdy";

    $$phProperties{'LckFile'} = $sLckFile;
    $$phProperties{'OutFile'} = $sOutFile;
    $$phProperties{'ErrFile'} = $sErrFile;
    $$phProperties{'EnvFile'} = $sEnvFile;
    $$phProperties{'RdyFile'} = $sRdyFile;

    if (!defined(MakePutTree($$phProperties{'IncomingDirectory'}, $sPutName, 0755, 1, \$sLocalError)))
    {
      $$psError = $sLocalError;
      SysReadWrite(\*STDIN, undef, $$phProperties{'ContentLength'}, undef); # Slurp up data to prevent a broken pipe.
      return 500;
    }

    ##################################################################
    #
    # Create a group lockfile and lock it. The purpose of the lock
    # is to prevent other instances of this script from writing to
    # any of the output files (.out, .err, .env, .rdy).
    #
    ##################################################################

    if (!open(LH, "> $sLckFile"))
    {
      $$psError = "File ($sLckFile) could not be opened ($!)";
      SysReadWrite(\*STDIN, undef, $$phProperties{'ContentLength'}, undef); # Slurp up data to prevent a broken pipe.
      return 500;
    }
    flock(LH, LOCK_EX);

    ##################################################################
    #
    # Make sure that none of the output files exist.
    #
    ##################################################################

    foreach my $sPutFile ($sOutFile, $sErrFile, $sEnvFile, $sRdyFile)
    {
      if (-e $sPutFile)
      {
        if (Yes($$phProperties{'OverwriteExistingFiles'}))
        {
          unlink($sPutFile);
        }
        else
        {
          $$psError = "File ($sPutFile) already exists";
          SysReadWrite(\*STDIN, undef, $$phProperties{'ContentLength'}, undef); # Slurp up data to prevent a broken pipe.
          flock(LH, LOCK_UN); close(LH); unlink($sLckFile); # Unlock, close, and remove the group lockfile.
          return 451;
        }
      }
    }

    ##################################################################
    #
    # Write the output files (.out, .err, .env, .rdy) to disk.
    #
    ##################################################################

    my (%hStreamLengths);

    $hStreamLengths{$sOutFile} = $sOutLength;
    $hStreamLengths{$sErrFile} = $sErrLength;
    $hStreamLengths{$sEnvFile} = $sEnvLength;

    foreach my $sPutFile ($sOutFile, $sErrFile, $sEnvFile, $sRdyFile)
    {
      if (!open(FH, "> $sPutFile"))
      {
        $$psError = "File ($sPutFile) could not be opened ($!)";
        SysReadWrite(\*STDIN, undef, $$phProperties{'ContentLength'}, undef); # Slurp up data to prevent a broken pipe.
        flock(LH, LOCK_UN); close(LH); unlink($sLckFile); # Unlock, close, and remove the group lockfile.
        return 500;
      }
      binmode(FH);
      flock(FH, LOCK_EX);
      if ($sPutFile eq $sRdyFile)
      {
        print FH "Version=", $$phProperties{'Version'}, $$phProperties{'Newline'};
        print FH "Jid=", $$phProperties{'JobId'}, $$phProperties{'Newline'};
        foreach my $sKey (sort(keys(%{$$phProperties{'GlobalConfigTemplate'}})))
        {
          print FH $sKey, "=", $$phProperties{$sKey}, $$phProperties{'Newline'};
        }
      }
      else
      {
        my $sByteCount = SysReadWrite(\*STDIN, \*FH, $hStreamLengths{$sPutFile}, \$sLocalError);
        if (!defined($sByteCount))
        {
          $$psError = $sLocalError;
          flock(FH, LOCK_UN); close(FH);
          flock(LH, LOCK_UN); close(LH); unlink($sLckFile); # Unlock, close, and remove the group lockfile.
          return 500;
        }
        if ($sByteCount != $hStreamLengths{$sPutFile})
        {
          $$psError = "Stream length ($hStreamLengths{$sPutFile}) does not equal number of bytes processed ($sByteCount) for output file ($sPutFile)";
          flock(FH, LOCK_UN); close(FH);
          flock(LH, LOCK_UN); close(LH); unlink($sLckFile); # Unlock, close, and remove the group lockfile.
          return 456;
        }
      }
      flock(FH, LOCK_UN); close(FH);
    }
    flock(LH, LOCK_UN); close(LH); unlink($sLckFile); # Unlock, close, and remove the group lockfile.
    if (Yes($$phProperties{'EnableJobQueues'}) && $$phProperties{'JqdQueueTag'} =~ /^$$phProperties{'CommonRegexes'}{'JqdQueueTag'}$/)
    {
      if (!UpdateQueue($phProperties, \$sLocalError))
      {
        $$psError = $sLocalError;
        return 500;
      }
      $$psError = "Success ($$phProperties{'JqdQueueTag'})";
    }
    else
    {
      $$psError = "Success";
    }
    return 200;
  }
  else
  {
    $$psError = "Invalid query string ($$phProperties{'QueryString'})";
    return 450;
  }
}


######################################################################
#
# SendResponse
#
######################################################################

sub SendResponse
{
  my ($phProperties) = @_;

  ####################################################################
  #
  # Send response header.
  #
  ####################################################################

  my ($sHandle, $sHeader, $sLength, $sReason, $sServer, $sStatus);

  $sHandle = $$phProperties{'ReturnHandle'};
  $sStatus = $$phProperties{'ReturnStatus'};
  $sReason = $$phProperties{'ReturnReason'};
  $sServer = $$phProperties{'ServerSoftware'} || "unknown";
  $sLength = (defined($sHandle)) ? -s $sHandle : ($$phProperties{'RequestMethod'} eq 'JOB') ? length($$phProperties{'JqdJobList'}) : 0;

  $sHeader  = "HTTP/1.1 $sStatus $sReason\r\n";
  $sHeader .= "Server: $sServer\r\n";
  $sHeader .= "Content-Type: application/octet-stream\r\n";
  $sHeader .= "Content-Length: $sLength\r\n";
  if ($$phProperties{'RequestMethod'} eq 'GET' && Yes($$phProperties{'EnableJobIds'}))
  {
    if (defined($$phProperties{'JobId'}) && $$phProperties{'JobId'} =~ /^$$phProperties{'CommonRegexes'}{'JobId'}$/)
    {
      $sHeader .= "Job-Id: $$phProperties{'JobId'}\r\n";
    }
  }
  if (exists($$phProperties{'DsvPayloadSignature'}) && defined($$phProperties{'DsvPayloadSignature'}))
  {
    $sHeader .= "WebJob-Payload-Signature: $$phProperties{'DsvPayloadSignature'}\r\n";
  }
  $sHeader .= "\r\n";

  syswrite(STDOUT, $sHeader, length($sHeader));

  ####################################################################
  #
  # Send content if any.
  #
  ####################################################################

  if (defined($sHandle))
  {
    SysReadWrite($sHandle, \*STDOUT, $sLength, undef);
    close($sHandle);
  }
  else
  {
    if ($$phProperties{'RequestMethod'} eq 'JOB')
    {
      syswrite(\*STDOUT, $$phProperties{'JqdJobList'}, $sLength);
    }
  }

  return $sLength;
}


######################################################################
#
# SetupHostAccessList
#
######################################################################

sub SetupHostAccessList
{
  my ($phProperties, $psError) = @_;

  ####################################################################
  #
  # Initialize the access list. This list is a hash that contains
  # client ID to IP address mappings. If the client passes all other
  # authentication tests, a lookup will be performed to determine if
  # the client's IP address is authorized. If it is, processing
  # continues as usual. Otherwise, the CGI script will return a 403
  # status code to the client, log an error message, and abort. If the
  # access list is empty or there is an error processing its file, all
  # access will be denied (i.e., this mechanism fails closed).
  #
  # The format of the list is key/value pair as follows:
  #
  #   <client_id> = <cidr>[,<cidr>]
  #
  #####################################################################

  %{$$phProperties{'AccessList'}} = (); # Initialize an empty list.

  my $sFile = $$phProperties{'NphConfigDirectory'} . "/" . "nph-webjob-hosts.access";

  if (open(AH, "< $sFile"))
  {
    while (my $sLine = <AH>)
    {
      $sLine =~ s/[\r\n]+$//; # Remove CRs and LFs.
      $sLine =~ s/#.*$//; # Remove comments.
      if ($sLine !~ /^\s*$/)
      {
        my ($sKey, $sValue) = ($sLine =~ /^([^=]*)=(.*)$/);
        $sKey =~ s/^\s+//; # Remove leading whitespace.
        $sKey =~ s/\s+$//; # Remove trailing whitespace.
        $sValue =~ s/^\s+//; # Remove leading whitespace.
        $sValue =~ s/\s+$//; # Remove trailing whitespace.
        if ($sKey =~ /^$$phProperties{'CommonRegexes'}{'ClientId'}$/)
        {
          $$phProperties{'AccessList'}{$sKey} = $sValue;
        }
      }
    }
    close(AH);
  }
  else
  {
    $$psError = "File ($sFile) could not be opened ($!)" if (defined($psError));
    return undef;
  }
}


######################################################################
#
# SysReadWrite
#
######################################################################

sub SysReadWrite
{
  my ($sReadHandle, $sWriteHandle, $sLength, $psError) = @_;

  ####################################################################
  #
  # Read/Write data, but discard data if write handle is undefined.
  #
  ####################################################################

  my ($sData, $sEOF, $sNRead, $sNProcessed, $sNWritten);

  for ($sEOF = $sNRead = $sNProcessed = 0; !$sEOF && $sLength > 0; $sLength -= $sNRead)
  {
    $sNRead = sysread($sReadHandle, $sData, ($sLength > 0x4000) ? 0x4000 : $sLength);
    if (!defined($sNRead))
    {
      $$psError = "Error reading from input stream ($!)" if (defined($psError));
      return undef;
    }
    elsif ($sNRead == 0)
    {
      $sEOF = 1;
    }
    else
    {
      if (defined($sWriteHandle))
      {
        $sNWritten = syswrite($sWriteHandle, $sData, $sNRead);
        if (!defined($sNWritten))
        {
          $$psError = "Error writing to output stream ($!)" if (defined($psError));
          return undef;
        }
      }
      else
      {
        $sNWritten = $sNRead;
      }
      $sNProcessed += $sNWritten;
    }
  }

  return $sNProcessed;
}


######################################################################
#
# TriggerExecuteCommandLine
#
######################################################################

sub TriggerExecuteCommandLine
{
  my ($phProperties, $psError) = @_;

  ####################################################################
  #
  # Make sure that required inputs are defined.
  #
  ####################################################################

  my %hLArgs =
  (
    'Hash' => $phProperties,
    'Keys' =>
    [
      'EnableLogging',
      'ExpandTriggerCommandLineRoutine',
      'OSClass',
    ],
  );
  if (!defined(VerifyHashKeys(\%hLArgs)))
  {
    $$psError = $hLArgs{'Error'} if (defined($phProperties));
    return undef;
  }

  ####################################################################
  #
  # Windows platforms are not currently supported.
  #
  ####################################################################

  if ($$phProperties{'OSClass'} eq "WINDOWS")
  {
    $$psError = "Triggers are not currently supported on Windows platforms";
    return undef;
  }

  ####################################################################
  #
  # Expand the trigger's command line. If the result is undefined or
  # null, abort.
  #
  ####################################################################

  my ($sLocalError);

  $$phProperties{'TriggerCommandLine'} = &{$$phProperties{'ExpandTriggerCommandLineRoutine'}}($phProperties, \$sLocalError);
  if (!defined($$phProperties{'TriggerCommandLine'}))
  {
    $$psError = $sLocalError;
    return undef;
  }

  if (!length($$phProperties{'TriggerCommandLine'}))
  {
    $$psError = "Command line is undefined or null";
    return undef;
  }

  ####################################################################
  #
  # Spawn a subprocess. Set the kid's process group. This should
  # isolate the kid from signals sent to his parent or grandparent
  # (i.e., this script or the server daemon, respectively).  Close
  # STDOUT. This should prevent the kid from interfering with the
  # original CGI connection (e.g., holding the socket open). Keep
  # STDERR open open so that errors can be caught in the server's
  # error log. Change to the root directory to prevent unmounting
  # issues, which could happen if a long-running trigger process was
  # specified.
  #
  ####################################################################

  my $sKidPid = fork();

  if (!defined($sKidPid))
  {
    $$psError = "Unable to spawn process ($!)";
    return undef;
  }
  else
  {
    if ($sKidPid == 0)
    {
      setpgrp(0, 0);
      close(STDOUT);
      chdir("/");
      $$phProperties{'TriggerPidLabel'} = "kid";
      $$phProperties{'TriggerPid'} = $$;
      $$phProperties{'TriggerState'} = "pulled";
      $$phProperties{'TriggerMessage'} = $$phProperties{'TriggerCommandLine'};
      if (Yes($$phProperties{'EnableLogging'}))
      {
        TriggerLogMessage($phProperties);
      }
      my $sKidReturn = system($$phProperties{'TriggerCommandLine'});
      my $sKidStatus = ($sKidReturn >> 8) & 0xff;
      my $sKidSignal = ($sKidReturn & 0x7f);
      my $sKidDumped = ($sKidReturn & 0x80) ? 1 : 0;
      if ($sKidStatus == 255)
      {
        $$phProperties{'TriggerState'} = "failed";
        $$phProperties{'TriggerMessage'} = "Unable to execute trigger command ($!)";
      }
      else
      {
        $$phProperties{'TriggerState'} = "reaped";
        $$phProperties{'TriggerMessage'} = "status($sKidStatus) signal($sKidSignal) coredump($sKidDumped)";
      }
      if (Yes($$phProperties{'EnableLogging'}))
      {
        TriggerLogMessage($phProperties);
      }
      exit($sKidStatus);
    }
  }

  return $sKidPid;
}


######################################################################
#
# TriggerLogMessage
#
######################################################################

sub TriggerLogMessage
{
  my ($phProperties) = @_;

  ####################################################################
  #
  # Marshall log arguments.
  #
  ####################################################################

  my (%hLArgs);

  $hLArgs{'LogEpoch'} = $$phProperties{'TriggerEpoch'};

  $hLArgs{'LogFields'} =
  [
    'JobId',
    'RequestMethod',
    'ClientId',
    'ClientFilename',
    'TriggerPidLabel',
    'TriggerPid',
    'TriggerState',
    'Message',
  ];

  $hLArgs{'LogValues'} =
  {
    'JobId'               => $$phProperties{'JobId'},
    'RequestMethod'       => $$phProperties{'RequestMethod'},
    'ClientId'            => $$phProperties{'ClientId'},
    'ClientFilename'      => $$phProperties{'ClientFilename'},
    'TriggerPidLabel'     => $$phProperties{'TriggerPidLabel'},
    'TriggerPid'          => $$phProperties{'TriggerPid'},
    'TriggerState'        => $$phProperties{'TriggerState'},
    'Message'             => $$phProperties{'TriggerMessage'},
  };

  $hLArgs{'LogFile'} = $$phProperties{'TriggerLogFile'};

  $hLArgs{'Newline'} = $$phProperties{'Newline'};

  $hLArgs{'RevertToStderr'} = 1;

  $hLArgs{'UseGmt'} = (Yes($$phProperties{'UseGmt'})) ? 1 : 0;

  ####################################################################
  #
  # Deliver log message.
  #
  ####################################################################

  LogNf1vMessage(\%hLArgs);

  1;
}


######################################################################
#
# UpdateQueue
#
######################################################################

sub UpdateQueue
{
  my ($phProperties, $psError) = @_;

  ####################################################################
  #
  # Make sure that required inputs are defined.
  #
  ####################################################################

  my %hLArgs =
  (
    'Hash' => $phProperties,
    'Keys' =>
    [
      'JqdQueueTag',
      'JqdQueueDirectory',
      'RequestMethod',
    ],
  );
  if (!defined(VerifyHashKeys(\%hLArgs)))
  {
    $$psError = $hLArgs{'Error'} if (defined($phProperties));
    return undef;
  }

  ####################################################################
  #
  # Determine the internal state directories that will be in play,
  # and prepare the this/next queue state filenames.
  #
  ####################################################################

  my ($sNextJobFile, $sNextJobState, $sProperty, $sPropertyValue, $sThisJobFile, $sThisJobState);

  if ($$phProperties{'RequestMethod'} eq "GET")
  {
    $sThisJobState = "sent";
    $sNextJobState = "open";
    $sProperty = "JobId";
  }
  else
  {
    $sThisJobState = "open";
    $sNextJobState = "done";
    $sProperty = "RdyFile";
  }
  $sPropertyValue = $$phProperties{$sProperty} || "";
  $sThisJobFile = $$phProperties{'JqdQueueDirectory'} . "/" . $sThisJobState . "/" . $$phProperties{'JqdQueueTag'};
  $sNextJobFile = $$phProperties{'JqdQueueDirectory'} . "/" . $sNextJobState . "/" . $$phProperties{'JqdQueueTag'};

  ####################################################################
  #
  # Lock the queue. This is an exclusive lock.
  #
  ####################################################################

  my (%hQueueLockArgs);

  %hQueueLockArgs =
  (
    'LockFile'  => $$phProperties{'JqdQueueDirectory'} . "/" . "change.lock",
  );

  if (!JqdLockFile(\%hQueueLockArgs))
  {
    $$psError = $hQueueLockArgs{'Error'};
    return undef;
  }

  ####################################################################
  #
  # Move the job file to the next queue state, and append additional
  # properties as necessary.
  #
  ####################################################################

  if (!rename($sThisJobFile, $sNextJobFile))
  {
    $$psError = "Unable to move the job ($sThisJobFile) to the next queue state ($!)";
    JqdUnlockFile(\%hQueueLockArgs);
    return undef;
  }

  if (defined($sProperty))
  {
    my %hLArgs =
    (
      'AppendToFile'   => 1,
      'File'           => $sNextJobFile,
      'Properties'     => { $sProperty => $$phProperties{$sProperty} },
      'Template'       => { $sProperty => undef },
      'VerifyValues'   => 0,
    );
    if (!KvpSetKvps(\%hLArgs))
    {
      $$psError = $hLArgs{'Error'};
      JqdUnlockFile(\%hQueueLockArgs);
      return undef;
    }
  }

  JqdUnlockFile(\%hQueueLockArgs);

  1;
}


######################################################################
#
# URLDecode
#
######################################################################

sub URLDecode
{
  my ($sData) = @_;

  $sData =~ s/\+/ /sg;
  $sData =~ s/%([0-9a-fA-F]{2})/pack('C', hex($1))/seg;

  return $sData;
}


######################################################################
#
# VerifyConversionValues
#
######################################################################

sub VerifyConversionValues
{
  my ($phConversionValues, $phConversionChecks, $psError) = @_;

  foreach my $sKey (sort(keys(%$phConversionChecks)))
  {
    if ($$phConversionValues{$sKey} !~ /^$$phConversionChecks{$sKey}$/)
    {
      $sKey =~ s/^(CID|CMD|IP|PID|SID)$/lc($1)/e; # Squash legacy tokens.
      $$psError = "Conversion value ($$phConversionValues{$sKey}) for corresponding specification (%$sKey) is not valid";
      return undef;
    }
  }

  1;
}


######################################################################
#
# VerifyRunTimeEnvironment
#
######################################################################

sub VerifyRunTimeEnvironment
{
  my ($phProperties, $phRequiredProperties, $psError) = @_;

  ####################################################################
  #
  # Make sure all required properties are defined and valid.
  #
  ####################################################################

  foreach my $sProperty (keys(%$phRequiredProperties))
  {
    my $sValue = $$phProperties{$sProperty};
    if (!defined($sValue) || $sValue !~ /^$$phRequiredProperties{$sProperty}$/)
    {
      $$psError = "$sProperty property ($sValue) is undefined or invalid";
      return undef;
    }
  }

  ####################################################################
  #
  # Make sure the config directory is readable.
  #
  ####################################################################

  if (!-d $$phProperties{'NphConfigDirectory'} || !-R _)
  {
    $$psError = "Config directory ($$phProperties{'NphConfigDirectory'}) does not exist or is not readable";
    return undef;
  }

  ####################################################################
  #
  # Make sure the dynamic directory is writable.
  #
  ####################################################################

  if (!-d $$phProperties{'DynamicDirectory'} || !-W _)
  {
    $$psError = "Dynamic directory ($$phProperties{'DynamicDirectory'}) does not exist or is not writable";
    return undef;
  }

  ####################################################################
  #
  # Make sure the logfiles directory is readable.
  #
  ####################################################################

  if (!-d $$phProperties{'LogfilesDirectory'} || !-R _)
  {
    $$psError = "Logfiles directory ($$phProperties{'LogfilesDirectory'}) does not exist or is not readable";
    return undef;
  }

  ####################################################################
  #
  # Make sure the profiles directory is readable.
  #
  ####################################################################

  if (!-d $$phProperties{'ProfilesDirectory'} || !-R _)
  {
    $$psError = "Profiles directory ($$phProperties{'ProfilesDirectory'}) does not exist or is not readable";
    return undef;
  }

  ####################################################################
  #
  # Make sure the incoming directory is writable.
  #
  ####################################################################

  if (!-d $$phProperties{'IncomingDirectory'} || !-W _)
  {
    $$psError = "Incoming directory ($$phProperties{'IncomingDirectory'}) does not exist or is not writable";
    return undef;
  }

  ####################################################################
  #
  # Make sure the spool directory is readable.
  #
  ####################################################################

  if (!-d $$phProperties{'SpoolDirectory'} || !-R _)
  {
    $$psError = "Spool directory ($$phProperties{'SpoolDirectory'}) does not exist or is not readable";
    return undef;
  }

  1;
}


######################################################################
#
# Yes
#
######################################################################

sub Yes
{
  return (defined($_[0]) && $_[0] =~ /^[Yy]$/) ? 1 : 0;
}

__END__

=pod

=head1 NAME

nph-webjob.cgi - Process webjob GET/JOB/PUT requests.

=head1 SYNOPSIS

    --- http.conf ---
    ...
    Include /var/webjob/config/apache/webjob.conf
    ...
    --- http.conf ---

    --- /var/webjob/config/apache/webjob.conf ---
    SetEnv WEBJOB_PROPERTIES_FILE /var/webjob/config/nph-webjob/nph-webjob.cfg

    ScriptAlias /cgi-client/ "/usr/local/webjob/cgi/cgi-client/"

    <Directory "/usr/local/webjob/cgi/cgi-client/">
      AllowOverride AuthConfig
      Options None
      Order allow,deny
      Allow from all
    </Directory>
    --- /var/webjob/config/apache/webjob.conf ---

=head1 DESCRIPTION

This utility implements a Common Gateway Interface (CGI), and its
purpose is to processes webjob GET/JOB/PUT requests.

The name of this utility must begin with the string "nph-".  This
naming convention, which stands for Non Parsed Headers, informs the
web server that the utility is responsible for generating the entire
HTTP response.  Non Parsed Headers are used by this utility because
they allow it to take full advantage of HTTP status code extensions.
This simplifies client-side error checking without giving up the
ability to generate application-specific responses.

The nph-webjob.cfg config file is used to control the CGI's runtime
behavior. To make use of it, set WEBJOB_PROPERTIES_FILE as
appropriate. Upon execution, the script looks for and processes the
file specified by this environment variable.

The cgi-client directory is where nph-webjob.cgi lives by default.
Access to this directory should be strictly controlled, and any
utilities placed in this directory should operate in the same security
context (or realm).  This utility is designed to read/write data
from/to the server's local disk on behalf of potentially hostile
clients.  Therefore, you'll want to maintain a tight rein to avoid any
unwanted results.

=head1 FILES

=over 5

=item B<BaseDirectory>/logfiles/nph-webjob.log

This utility writes log messages to nph-webjob.log (see
nph-webjob.cfg).  This file contains the following fields separated by
whitespace in the given order: DateTime, JobId, RemoteUser,
RemoteAddress, RequestMethod, ClientId, ClientFilename, ContentLength,
ServerContentLength, Duration, ReturnStatus, and ErrorMessage.  A
single hyphen, '-', is used to denote the fact that a particular field
had an empty or undefined value.  The ErrorMessage field is set off
from the other fields with a double hyphen '--' because it uses a free
form text format that can include whitespace.

=item B<BaseDirectory>/logfiles/nph-webjob-hook.err

This utility captures hook-related output on stderr and writes it to
nph-webjob-hook.err.  This file is a bit bucket for dirty hook
commands.  It is also useful as a debugging aid when creating new hook
commands.

=item B<BaseDirectory>/logfiles/nph-webjob-hook.log

This utility writes hook-related log messages to nph-webjob-hook.log.
This file contains the following fields separated by whitespace in the
given order: Date, Time, JobId, RequestMethod, ClientId,
ClientFilename, PidLabel, Pid, State, and Message.  A single hyphen,
'-', is used to denote the fact that a particular field had an empty
or undefined value.  The Message field is set off from the other
fields with a double hyphen '--' because it uses a free form text
format that can include whitespace.

=item B<BaseDirectory>/logfiles/nph-webjob-hook.out

This utility captures hook-related output on stdout and writes it to
nph-webjob-hook.out.  This file is a bit bucket for dirty hook
commands.  It is also useful as a debugging aid when creating new hook
commands.

=item B<BaseDirectory>/logfiles/nph-webjob-trigger.log

This utility writes trigger-related log messages to
nph-webjob-trigger.log.  This file contains the following fields
separated by whitespace in the given order: Date, Time, JobId,
RequestMethod, ClientId, ClientFilename, PidLabel, Pid, State, and
Message.  A single hyphen, '-', is used to denote the fact that a
particular field had an empty or undefined value.  The Message field
is set off from the other fields with a double hyphen '--' because it
uses a free form text format that can include whitespace.

=back

=head1 AUTHOR

Klayton Monroe

=head1 SEE ALSO

nph-config.cgi(1), webjob(1)

=head1 LICENSE

All documentation and code are distributed under same terms and
conditions as B<WebJob>.

=cut
